import tskit, pyslim

"""
Takes an old tree sequence and update the metadata *without* properly updating
the top-level metadata.
"""

ts = tskit.load("recipe_WF.v3.5.trees")
tables = ts.dump_tables()

tables.populations.clear()
tables.populations.metadata_schema = pyslim.slim_metadata_schemas['population']
for p in ts.populations():
    tables.populations.append(p)

tables.individuals.clear()
tables.individuals.metadata_schema = pyslim.slim_metadata_schemas['individual']
d = pyslim.default_slim_metadata("individual")
for i in ts.individuals():
    d.update(i.metadata)
    ii = i.replace(metadata=d)
    tables.individuals.append(ii)

tables.mutations.clear()
tables.mutations.metadata_schema = pyslim.slim_metadata_schemas['mutation']
d = pyslim.default_slim_metadata("mutation")
for m in ts.mutations():
    tables.mutations.append(m)

ts = tables.tree_sequence()
ts.dump("recipe_WF.v3.5_and_v3.6.trees")



"""
Common code for the pyslim test cases.
"""
import os
import json
import random
import warnings

import pyslim
import tskit
import msprime
import pytest
import numpy as np


class PyslimTestCase:
    '''
    Base class for test cases in pyslim.
    '''

    def verify_haplotype_equality(self, ts, slim_ts):
        assert ts.num_sites == slim_ts.num_sites
        for j, v1, v2 in zip(range(ts.num_sites), ts.variants(),
                             slim_ts.variants()):
            g1 = [v1.alleles[x] for x in v1.genotypes]
            g2 = [v2.alleles[x] for x in v2.genotypes]
            assert np.array_equal(g1, g2)


    def assertTablesEqual(self, t1, t2, label=''):
        # make it easy to see what's wrong
        if hasattr(t1, "metadata_schema"):
            if t1.metadata_schema != t2.metadata_schema:
                print(f"{label} :::::::::: t1 ::::::::::::")
                print(t1.metadata_schema)
                print(f"{label} :::::::::: t2 ::::::::::::")
                print(t2.metadata_schema)
            assert t1.metadata_schema == t2.metadata_schema
        if t1.num_rows != t2.num_rows:
            print(f"{label}: t1.num_rows {t1.num_rows} != {t2.num_rows} t2.num_rows")
        for k, (e1, e2) in enumerate(zip(t1, t2)):
            if e1 != e2:
                print(f"{label} :::::::::: t1 ({k}) ::::::::::::")
                print(e1)
                print(f"{label} :::::::::: t2 ({k}) ::::::::::::")
                print(e2)
            assert e1 == e2
        assert t1.num_rows == t2.num_rows
        assert t1 == t2

    def assertMetadataEqual(self, t1, t2):
        # check top-level metadata, first the parsed version:
        assert t1.metadata_schema == t2.metadata_schema
        assert t1.metadata == t2.metadata
        # and now check the underlying bytes
        # TODO: use the public interface if https://github.com/tskit-dev/tskit/issues/832 happens
        md1 = t1._ll_tables.metadata
        md2 = t2._ll_tables.metadata
        assert md1 == md2

    def verify_trees_equal(self, ts1, ts2):
        # check that trees are equal by checking MRCAs between randomly
        # chosen nodes with matching slim_ids
        random.seed(23)
        assert ts1.sequence_length == ts2.sequence_length
        if isinstance(ts1, tskit.TableCollection):
            ts1 = ts1.tree_sequence()
        if isinstance(ts2, tskit.TableCollection):
            ts2 = ts2.tree_sequence()
        map1 = {}
        for j, n in enumerate(ts1.nodes()):
            if n.metadata is not None:
                map1[n.metadata['slim_id']] = j
        map2 = {}
        for j, n in enumerate(ts2.nodes()):
            if n.metadata is not None:
                map2[n.metadata['slim_id']] = j
        assert set(map1.keys()) == set(map2.keys())
        print(ts1)
        print(map1)
        print(ts2)
        print(map2)
        sids = list(map1.keys())
        for sid in sids:
            n1 = ts1.node(map1[sid])
            n2 = ts2.node(map2[sid])
            assert n1.time == n2.time
            assert n1.metadata == n2.metadata
            i1 = ts1.individual(n1.individual)
            i2 = ts2.individual(n2.individual)
            if i1.metadata != i2.metadata:
                print("i1: ", i1.metadata)
                print("i2: ", i2.metadata)
            assert i1.metadata == i2.metadata
        for _ in range(10):
            pos = random.uniform(0, ts1.sequence_length)
            t1 = ts1.at(pos)
            t2 = ts2.at(pos)
            for _ in range(10):
                a, b = random.choices(sids, k=2)
                print(a, b, map1[a], map1[b], map2[a], map2[b])
                print(t1)
                print('a', t1.time(map1[a]))
                print('b', t1.time(map1[b]))
                print('1', t1.tmrca(map1[a], map1[b]))
                print('2', t2.tmrca(map2[a], map2[b]))
                assert t1.tmrca(map1[a], map1[b]) == t2.tmrca(map2[a], map2[b])

    def assertTableCollectionsEqual(self, t1, t2,
            skip_provenance=False, check_metadata_schema=True,
            reordered_individuals=False):
        if isinstance(t1, tskit.TreeSequence):
            t1 = t1.dump_tables()
        if isinstance(t2, tskit.TreeSequence):
            t2 = t2.dump_tables()
        t1_samples = [(n.metadata['slim_id'], j) for j, n in enumerate(t1.nodes) if (n.flags & tskit.NODE_IS_SAMPLE)]
        t1_samples.sort()
        t2_samples = [(n.metadata['slim_id'], j) for j, n in enumerate(t2.nodes) if (n.flags & tskit.NODE_IS_SAMPLE)]
        t2_samples.sort()
        t1.simplify([j for (_, j) in t1_samples], record_provenance=False)
        t2.simplify([j for (_, j) in t2_samples], record_provenance=False)
        if skip_provenance is True:
            t1.provenances.clear()
            t2.provenances.clear()
        if skip_provenance == -1:
            assert t1.provenances.num_rows + 1 == t2.provenances.num_rows
            t2.provenances.truncate(t1.provenances.num_rows)
            assert t1.provenances.num_rows == t2.provenances.num_rows
        if check_metadata_schema:
            # this is redundant now, but will help diagnose if things go wrong
            assert t1.metadata_schema.schema == t2.metadata_schema.schema
            assert t1.populations.metadata_schema.schema == t2.populations.metadata_schema.schema
            assert t1.individuals.metadata_schema.schema == t2.individuals.metadata_schema.schema
            assert t1.nodes.metadata_schema.schema == t2.nodes.metadata_schema.schema
            assert t1.edges.metadata_schema.schema == t2.edges.metadata_schema.schema
            assert t1.sites.metadata_schema.schema == t2.sites.metadata_schema.schema
            assert t1.mutations.metadata_schema.schema == t2.mutations.metadata_schema.schema
            assert t1.migrations.metadata_schema.schema == t2.migrations.metadata_schema.schema
        if not check_metadata_schema:
            # need to pull out metadata to compare as dicts before zeroing the schema
            m1 = t1.metadata
            m2 = t2.metadata
            ms = tskit.MetadataSchema(None)
            for t in (t1, t2):
                t.metadata_schema = ms
                t.populations.metadata_schema = ms
                t.individuals.metadata_schema = ms
                t.nodes.metadata_schema = ms
                t.edges.metadata_schema = ms
                t.sites.metadata_schema = ms
                t.mutations.metadata_schema = ms
                t.migrations.metadata_schema = ms
            t1.metadata = b''
            t2.metadata = b''
            assert m1 == m2
        if reordered_individuals:
            ind1 = {i.metadata['pedigree_id']: j for j, i in enumerate(t1.individuals)}
            ind2 = {i.metadata['pedigree_id']: j for j, i in enumerate(t2.individuals)}
            for pid in ind1:
                if not pid in ind2:
                    print("not in t2:", ind1[pid])
                assert pid in ind2
                if t1.individuals[ind1[pid]] != t2.individuals[ind2[pid]]:
                    print("t1:", t1.individuals[ind1[pid]])
                    print("t2:", t2.individuals[ind2[pid]])
                assert t1.individuals[ind1[pid]] == t2.individuals[ind2[pid]]
            for pid in ind2:
                if not pid in ind1:
                    print("not in t1:", ind2[pid])
                assert pid in ind1
            t1.individuals.clear()
            t2.individuals.clear()
        # go through one-by-one so we know which fails
        self.assertTablesEqual(t1.populations, t2.populations, "populations")
        self.assertTablesEqual(t1.individuals, t2.individuals, "individuals")
        self.assertTablesEqual(t1.nodes, t2.nodes, "nodes")
        self.assertTablesEqual(t1.edges, t2.edges, "edges")
        self.assertTablesEqual(t1.sites, t2.sites, "sites")
        self.assertTablesEqual(t1.mutations, t2.mutations, "mutations")
        self.assertTablesEqual(t1.migrations, t2.migrations, "migrations")
        self.assertTablesEqual(t1.provenances, t2.provenances, "provenances")
        self.assertMetadataEqual(t1, t2)
        assert t1.sequence_length == t2.sequence_length
        if t1.reference_sequence.data != t2.reference_sequence.data:
            print(t1.reference_sequence.data, " != ", t2.reference_sequence.data)
        assert t1.reference_sequence.data == t2.reference_sequence.data
        # time units are not passed on: https://github.com/tskit-dev/msprime/issues/1951
        # if t1.time_units != t2.time_units:
        #     print(t1.time_units, " != ", t2.time_units)
        # assert t1.time_units == t2.time_units
        # assert t1 == t2

    def do_recapitate(self, ts, *args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', msprime.TimeUnitsMismatchWarning)
            recap = pyslim.recapitate(ts, *args, **kwargs)
        return recap




"""specify which tests to run on each of the recipes in ./test_recipes"""

import os

# possible attributes to simulation scripts are
#  WF, nonWF
#  nucleotides
#  mutation_spectrum: writes out info file on mutation spectrum
#  everyone: records everyone ever
#  pedigree: writes out accompanying info file containing the pedigree
#  remembered_early: remembering and saving the ts happens during early
#  multipop: has more than one population
# All files are of the form `tests/test_recipes/{key}`
recipe_specs = {
    "recipe_nonWF.slim":                       {"nonWF": True, "pedigree": True},
    "recipe_long_nonWF.slim":                  {"nonWF": True, "long": True},
    "recipe_WF.slim":                          {"WF": True, "pedigree": True},
    "recipe_long_WF.slim":                     {"WF": True, "long": True},
    "recipe_WF_migration.slim":                {"WF": True, "pedigree": True, "multipop": True},
    "recipe_nucleotides_WF.slim":              {"WF": True, "pedigree": True, "nucleotides": True},
    "recipe_nucleotides_nonWF.slim":           {"nonWF": True, "pedigree": True, "nucleotides": True},
    "recipe_nucleotides_plus_others.slim":     {"WF": True, "pedigree": True, "nucleotides": True, "non-nucleotides": True, "adds_mutations": True},
    "recipe_long_nucleotides.slim":            {"WF": True, "nucleotides": True, "long": True},
    "recipe_mutation_spectrum.slim":           {"WF": True, "nucleotides": True, "mutation_spectrum": True},
    "recipe_roots.slim":                       {"WF": True, "pedigree": True, "long": True},
    "recipe_nonWF_selfing.slim":               {"nonWF": True, "pedigree": True},
    "recipe_nonWF_nonstacked.slim":            {"nonWF": True, "nonstacked": True},
    "recipe_init_mutated_WF.slim":             {"WF": True, "init_mutated": True},
    "recipe_init_mutated_nonWF.slim":          {"nonWF": True, "init_mutated": True},
    "recipe_with_metadata.slim":               {"user_metadata": True},
    "recipe_retain_everyone_WF_late.slim":     {"WF": True, "pedigree": True, "retained": True},
    "recipe_retain_sometimes_WF_late.slim":    {"WF": True, "pedigree": True, "retained": True},
    "recipe_retain_unary_WF_late.slim":        {"WF": True, "pedigree": True, "retained": True, "retainCoalescentOnly": False},
    "recipe_retain_everyone_nonWF_late.slim":  {"nonWF": True, "pedigree": True, "retained": True},
    "recipe_retain_sometimes_nonWF_late.slim": {"nonWF": True, "pedigree": True, "retained": True},
    "recipe_retain_unary_nonWF_late.slim":     {"nonWF": True, "pedigree": True, "retained": True, "retainCoalescentOnly": False},
    "recipe_remember_and_retain.slim":         {"nonWF": True, "pedigree": True, "retained": True},
}

for x in ("first", "early", "late"):
    for y in ("first", "early", "late"):
        for t in ("WF", "nonWF"):
            d = {"pedigree": True}
            d[t] = True
            if y != "late":
                # "remembered_early" or "remembered_first"
                d[f"remembered_{y}"] = True
            if x != "early":
                # "begun_late" or "begun_first"
                d[f"begun_{x}"] = True
            recipe_specs[f"recipe_{t}_{x}_{y}.slim"] = d

for y in ("first", "early", "late"):
    for t in ("WF", "nonWF"):
        d = {"pedigree": True, "everyone": True}
        d[t] = True
        if y != "late":
            # "remembered_early" or "remembered_first"
            d[f"remembered_{y}"] = True
        recipe_specs[f"recipe_record_everyone_{t}_{y}.slim"] = d


def recipe_eq(*keys, exclude=None):
    """
    Return an iterator over those recipes whose spec contains the specified keys.
    If key is empty, return all of them.
    If exclude is given exclude recipes with the specified keys.
    """
    if isinstance(exclude, str) or exclude is None:
        exclude = [exclude]
    return (
        k for k, v in recipe_specs.items()
        if (all(u not in v for u in exclude) and all(kk in v for kk in keys))
    )

# These SLiM scripts read in an existing trees file; the "input" gives a key in the
# recipe_specs array that will produce a "ts" file suitable for input
restarted_recipe_specs = {
    "restart_nucleotides_WF.slim":   {"WF": True, "nucleotides": True, "no_op": True, "input": "recipe_nucleotides_WF.slim"},
    "restart_nucleotides_nonWF.slim":   {"nonWF": True, "nucleotides": True, "no_op": True, "input": "recipe_nucleotides_nonWF.slim"},
    "restart_and_run_WF.slim":    {"WF": True, "input": "recipe_init_mutated_WF.slim"},
    "restart_and_run_nonWF.slim": {"nonWF": True, "input": "recipe_init_mutated_nonWF.slim"},
    "restart_and_remove_subpop_nonWF.slim":    {"nonWF": True, "input": "recipe_init_mutated_nonWF.slim", "remove_subpop": True},
}
for t in ("WF", "nonWF"):
    # recipes that read in and write out immediately ("no_op")
    value = {t: True, "no_op": True, "input": f"recipe_{t}.slim"}
    restarted_recipe_specs[f'restart_{t}.slim'] = value

def restarted_recipe_eq(*keys):
    """
    Return an iterator over a tuple of (restarted_recipe_name, input_recipe_name) giving
    those restarted recipes whose spec contains the specified keys.
    If key is empty, return all of them.
    """
    for k, v in restarted_recipe_specs.items():
        if all(kk in v for kk in keys):
            assert v["input"] in recipe_specs  # we need a valid input recipe
            yield (k, v["input"])




# Keywords: Python, tree-sequence recording, tree sequence recording

import tskit, pyslim
import numpy as np
import matplotlib.pyplot as plt

# Load the .trees file
ts = tskit.load("recipe_17.10_decap.trees")    # no simplify!

# Calculate tree heights, giving uncoalesced sites the maximum time
def tree_heights(ts):
    heights = np.zeros(ts.num_trees + 1)
    for tree in ts.trees():
        if tree.num_roots > 1:  # not fully coalesced
            heights[tree.index] = ts.metadata['SLiM']['tick']
        else:
            children = tree.children(tree.root)
            real_root = tree.root if len(children) > 1 else children[0]
            heights[tree.index] = tree.time(real_root)
    heights[-1] = heights[-2]  # repeat the last entry for plotting with step
    return heights

# Plot tree heights before recapitation
breakpoints = list(ts.breakpoints())
heights = tree_heights(ts)
plt.step(breakpoints, heights, where='post')
plt.show()

# Recapitate!
recap = pyslim.recapitate(ts, ancestral_Ne=1e5, recombination_rate=3e-10, random_seed=1)
recap.dump("recipe_17.10_recap.trees")

# Plot the tree heights after recapitation
breakpoints = list(recap.breakpoints())
heights = tree_heights(recap)
plt.step(breakpoints, heights, where='post')
plt.show()




from collections import namedtuple
import os
import pytest
import random
import subprocess

from filelock import FileLock
import json
import tskit, msprime

from .recipe_specs import recipe_specs

class Outfiles:
    """
    Simple wrapper for a list of Outfile items: these specify a file path that SLiM can
    use to output a file, the SLiM variable name containing that path, for use in the
    recipe, a function to apply to the path to create an object for the unit test to
    access and a "name" (key) in the dictionary to access that object in the unit test.
    """
    Outfile = namedtuple("Outfile", "path, slim_name, post_process, key")

    @staticmethod
    def read_dictionary(fname):
        # reads in the json-written dictionary
        assert os.path.isfile(fname)
        with open(fname, "r") as f:
            out = json.load(f)
        return out

    @staticmethod
    def parse_pedigree_info(fname):
        # returns a dictionary whose keys are SLiM individual IDs, and whose values
        # are dictionaries with two entries:
        # - 'parents' is the SLiM IDs of the parents
        # - 'age' is a dictionary whose keys are tuples (SLiM generation, stage)
        #   and whose values are ages (keys not present are ones the indivdiual was
        #   not alive for)
        assert os.path.isfile(fname)
        out = {}
        with open(fname, 'r') as f:
            header = f.readline().split()
            assert header == ['generation', 'stage', 'individual', 'age', 'parent1', 'parent2']
            for line in f:
                gen, stage, ind, age, p1, p2 = line.split()
                gen = int(gen)
                ind = int(ind)
                age = int(age)
                parents = tuple([int(p) for p in (p1, p2) if p != "-1"])
                if ind not in out:
                    out[ind] = {
                            "parents" : parents,
                            "age" : {}
                            }
                else:
                    for p in parents:
                        assert p in out[ind]['parents']
                out[ind]['age'][(gen, stage)] = age
        return out

    def __init__(self, out_dir):
        # Note: the 'key' below cannot match a 'key' in recipe_specs
        self._outfiles = [
            self.Outfile(
                path=os.path.join(out_dir, "out.trees"),
                slim_name="TREES_FILE",  # The var containing the path name for SLiM output
                post_process=tskit.load,  # function applied on path to create returned obj
                key="ts",  # The key to use for the post-processed item in the returned dict
            ),
            self.Outfile(
                path=os.path.join(out_dir, "out.pedigree"),
                slim_name="PEDIGREE_FILE",
                post_process=self.parse_pedigree_info,
                key="info",
            ),
            self.Outfile(
                path=os.path.join(out_dir, "out_mutations.json"),
                slim_name="MUTATIONS_FILE",
                post_process=self.read_dictionary,
                key="mutation_info",
            ),
        ]

    def __getitem__(self, index):
        return self._outfiles[index]

    def __len__(self, index):
        return len(self._outfiles)
    
    def results(self):
        res = {"path": {}}
        for o in self._outfiles:
            res["path"][o.key] = o.path
            if os.path.isfile(o.path):
                assert o.key not in res
                res[o.key] = o.post_process(o.path) 
        return res


class MultiOutfiles(Outfiles):
    
    def __init__(self, out_dir, species):
        # Note: the 'key' below cannot match a 'key' in recipe_specs
        self._outfiles = []
        for sp in species:
            self._outfiles.extend([
                self.Outfile(
                    path=os.path.join(out_dir, f"{sp}_out.trees"),
                    slim_name=f"TREES_FILE_{sp}",  # The var containing the path name for SLiM output
                    post_process=tskit.load,  # function applied on path to create returned obj
                    key=f"ts_{sp}",  # The key to use for the post-processed item in the returned dict
                ),
                self.Outfile(
                    path=os.path.join(out_dir, f"{sp}_out.pedigree"),
                    slim_name=f"PEDIGREE_FILE_{sp}",
                    post_process=self.parse_pedigree_info,
                    key=f"info_{sp}",
                ),
                self.Outfile(
                    path=os.path.join(out_dir, f"{sp}_out_mutations.json"),
                    slim_name=f"MUTATIONS_FILE_{sp}",
                    post_process=self.read_dictionary,
                    key=f"mutation_info_{sp}",
                ),
            ])


def run_slim(recipe, out_dir, recipe_dir="test_recipes", species=None, **kwargs):
    """
    Run the recipe, present in the recipe_dir, outputting to files in out_dir
    kwargs are passed as variables to SLiM (in addition to the outfiles)
    """
    script_dir = os.path.dirname(os.path.realpath(__file__))
    full_recipe = os.path.abspath(os.path.join(script_dir, recipe_dir, recipe))
    if not os.path.isfile(full_recipe):
        raise RuntimeError(f"{full_recipe} cannot be found")
    assert os.path.isdir(out_dir)  # should have been generated by the calling function
    if species is None:
        outfiles = Outfiles(out_dir)
    else:
        outfiles = MultiOutfiles(out_dir, species)
    slim_vars = []
    for o in outfiles:
        if o.slim_name != "":
            # for happy windows filepaths
            tmp_str = o.path.replace('\\', '\\\\')
            slim_vars += ["-d", f"{o.slim_name}=\"{tmp_str}\""]
    for k in kwargs:
        x = kwargs[k]
        if x is not None:
            if isinstance(x, str) and x[:10] != 'Dictionary':
                # for happy windows filepaths
                x = x.replace('\\', '\\\\')
                x = f"'{x}'"
            if isinstance(x, bool):
                x = 'T' if x else 'F'
            slim_vars += ["-d", f"{k}={x}"]
    cmd = ["slim", "-s", "23"] + slim_vars + [full_recipe]
    print(f"Running {' '.join(cmd)}, outputting errors etc. to '{out_dir}/SLiM_run_output.log'")
    with open(os.path.join(out_dir, "SLiM_run_output.log"), "w") as out:
        retval = subprocess.call(cmd, stderr=subprocess.STDOUT, stdout=out)
    if retval != 0:
        errmsg = f"Could not run {' '.join(cmd)}"
        errline = False
        with open(os.path.join(out_dir, "SLiM_run_output.log"), "r") as err:
            for line in err:
                if line[:5] == "ERROR" or line[:5] == "Error":
                    errline = True
                    errmsg += "\n"
                if errline and len(line.strip()) > 0:
                    errmsg += line
        raise RuntimeError(errmsg)
    return outfiles.results()


def _obj_to_eidos(x):
    if isinstance(x, str):
        out = '"' + x + '"'
    elif isinstance(x, dict):
        out = _dict_to_eidos(x)
    elif x is None:
        out = 'NULL'
    else:
        out = x
    return out


def _dict_to_eidos(d):
    txt = "Dictionary(" + ", ".join([f"\"{a}\", {_obj_to_eidos(b)}" for a, b in d.items()]) + ")"
    return txt


class HelperFunctions:
    @classmethod
    def run_slim_restart(cls, in_ts, recipe, out_dir, subpop_map=None, **kwargs):
        # Saves out the tree sequence to a trees file and run the SLiM recipe
        # on it, saving to files in out_dir.
        infile = os.path.join(out_dir, "in.trees")
        in_ts.dump(infile)
        # for happy windows filepaths       
        infile_str = infile.replace('\\', '\\\\')
        kwargs['TREES_IN'] = infile
        if 'STAGE' not in kwargs:
            kwargs['STAGE'] = in_ts.metadata['SLiM']['stage']
        if subpop_map is not None:
            # subpopMap argument has entries like { "p1" : 2 },
            # meaning that when loaded in, subpop p1 will be the second row
            # in the population table
            spm = _dict_to_eidos(subpop_map)
            kwargs['SUBPOP_MAP'] = spm
        results = run_slim(recipe, out_dir, **kwargs)
        return results["ts"]

    @classmethod
    def run_multispecies_slim_restart(cls, in_ts, recipe, out_dir, subpop_map=None, **kwargs):
        # in_ts should now be a dictionary with keys corresponding to species names
        infiles = {}
        for sp in in_ts:
            infiles[sp] = os.path.join(out_dir, f"{sp}_in.trees")
            in_ts[sp].dump(infiles[sp])
        infiles_str_dict = {}
        for k, v in infiles.items():
            # for happy windows filepaths
            infiles_str_dict[k] = v.replace('\\', '\\\\')        
        kwargs['TREES_IN'] = _dict_to_eidos(infiles_str_dict)
        if 'STAGE' not in kwargs:
            kwargs['STAGE'] = in_ts[sp].metadata['SLiM']['stage']
        if subpop_map is not None:
            # subpopMap argument has entries like { "p1" : 2 },
            # meaning that when loaded in, subpop p1 will be the second row
            # in the population table
            kwargs['SUBPOP_MAP'] = _dict_to_eidos(subpop_map)
        results = run_slim(recipe, out_dir, species=list(in_ts.keys()), **kwargs)
        out = { sp: results[f"ts_{sp}"] for sp in in_ts }
        return out

    @classmethod
    def run_msprime_restart(cls, in_ts, out_dir, sex=None, WF=False):
        out_ts = cls.run_slim_restart(
            in_ts,
            "restart_msprime.slim",  # This is standard script defined in test_recipes
            out_dir,
            WF=WF,
            SEX=sex,
            L=int(in_ts.sequence_length)
        )
        return out_ts

    @staticmethod
    def get_msprime_examples():
        # NOTE: we use DTWF below to avoid rounding of floating-point times
        # that occur with a continuous-time simulator
        seed = 6
        mutrate = 0.01
        for n in [2, 20]:
            for recrate in [0.0, 0.01]:
                ts = msprime.sim_ancestry(
                        n,
                        recombination_rate=recrate,
                        population_size=10,
                        sequence_length=200,
                        random_seed=seed,
                        model="dtwf"
                )
                yield ts
                mts = msprime.sim_mutations(
                        ts,
                        model=msprime.SLiMMutationModel(type=0),
                        rate=mutrate
                )
                yield mts
                seed += 1
                demography = msprime.Demography()
                demography.add_population(name="A", initial_size=20)
                demography.add_population(name="B", initial_size=30)
                demography.add_population(name="C", initial_size=10)
                demography.add_population_split(
                        time=5,
                        derived=["A", "B"],
                        ancestral="C"
                )
                ts = msprime.sim_ancestry(
                        {"A": n, "B": n},
                        demography=demography,
                        recombination_rate=recrate,
                        sequence_length=250,
                        random_seed=seed,
                        model="dtwf"
                )
                yield ts
                seed += 1


@pytest.fixture
def helper_functions():
    return HelperFunctions


@pytest.fixture(scope="session", params=recipe_specs.keys())
def recipe(request, tmp_path_factory, worker_id):
    """
    Most logic in this fixture is to check whether we are running as a single proc, or as a worker.
    If a worker, and this is the first run, we need to lock the process until the simulation has finished.
    """
    recipe_name = request.param
    if worker_id == "master":
        out_dir = tmp_path_factory.getbasetemp() / recipe_name
    else:
        # get the temp directory shared by all workers, so that simulation files
        # are shared between workers
        out_dir = tmp_path_factory.getbasetemp().parent / recipe_name
    with FileLock(str(out_dir) + ".lock"):
        if out_dir.is_dir():
            ret = Outfiles(out_dir).results()
        else:
            os.makedirs(out_dir, exist_ok=True)
            ret = run_slim(recipe=recipe_name, out_dir=out_dir)
    
    ret.update(recipe_specs[recipe_name])
    return ret



import msprime
import tskit
import warnings
import numpy as np

from .methods import *


def _in_location_bin(locations, x0, x1, y0, y1):
    '''
    Takes the locations of individuals and tests if each is within [x0, x1) and
    [y0, y1). Returns a list the same length as the number of individuals with
    True if the corresponding individual is within the bounds and False
    otherwise.  This is a helper function for `population_size`, and may change
    or be removed in the future.

    :param numpy.ndarray locations: An n x 3 array with rows individuals, the first
        column the x coordinates, and the second column the y coordinates.
    :param float x0: The lower x coordinate boundary.
    :param float x1: The upper x coordinate boundary.
    :param float y0: The lower y coordinate boundary.
    :param float y1: The upper y coordinate boundary.
    '''
    return np.logical_and(
            np.logical_and(
                locations[:,0] < x1,
                locations[:,0] >= x0),
            np.logical_and(
                locations[:,1] < y1,
                locations[:,1] >= y0
            )
        )


def _average_time_alive(birth_times, death_times, t0, t1):
    '''
    Calculates the average population size in the time interval [`t0`, `t1`)
    from birth times and death times of individuals. This is a helper function for
    `population_size`, and may change or be removed in the future.

    'Average population size' is the sum over all individuals of the length of
    time each individual was alive in [`t0`, `t1`), divided by `t1 - t0`. For
    integer t0 and t1, this is equivalent to the average over all integer time
    steps in [t0, t1) of the number of individuals alive at that time step;
    however, this is not restricted to integer times.

    Time goes bacwards, so t1 is further in the past than t0 and birth times
    are greater than death times. Birth times and death times should already
    include an offset for the stage that individuals are remembered at. See
    `population_size` code.

    For example, suppose t0 = 1, t1 = 4, and there are four individuals with
    birth times [3, 1, 5, 4] and death times [0, 1, 2, 4]. The first individual
    is alive for three time steps in [t0, t1): 1, 2, and 3; the second
    individual for one time step: 1; the third individual for two time steps: 2
    and 3, and the fourth individual for zero (since the time interval is open
    on the right and it was born at t1). So the average population size is (3 +
    1 + 2 + 0)/(4-1) = 2.

    :param numpy.ndarray birth_times: Birth times of individuals.
    :param numpy.ndarray death_times: Death times of individuals. Should be the same
        length as `birth_times`. Death times are less than or equal to birth times.
    :param float t0: Lower time endpoint.
    :param float t1: Upper time endpoint. t1 is greater than t0.

    '''
    # length of the segment of time in [t0, t1) that the individual
    # was alive for, i.e., of the intersection with [birth, death].
    a = np.maximum(0, np.minimum(t1, 1 + birth_times) -  np.maximum(t0, death_times))
    return sum(a)/(t1-t0)


def population_size(ts, x_bins, y_bins, time_bins, stage='late', remembered_stage=None):
    '''
    Calculates the population size in each of the spatial bins defined by grid lines
    at ``x_bins`` and ``y_bins``, averaged over each of the time intervals separated by
    ``time_bins``. To obtain actual (census) sizes, the tree sequence must contain
    all individuals alive, e.g., from a SLiM simulation with all individuals
    permanently remembered.

    With ``nx``, ``ny`` and ``nt`` the number of bins in the ``x``, ``y`` and time directions
    (so, ``nx`` is one less than the length of ``x_bins``),
    returns a 3-d array with dimensions ``(nx, ny, nt)``.
    The ``(i,j,k)``th element of the array is the average number of
    individuals with ``x`` coordinate in the half-open interval ``[x_bins[i], x_bins[i + 1])``
    and ``y`` coordinate in the half-open interval ``[y_bins[i], y_bins[i + 1])``,
    averaged across all times in the half-open interval ``[time_bins[k], time_bins[k + 1])``.

    For integer endpoints of the time bins, this average is equivalent to
    recording the number of indivduals that are alive at each time in the time
    interval and have location in the relevant location bin, then taking the
    mean of these recorded population sizes.

    :param tskit.TreeSequence ts: The tree sequence to calculate population size from.
    :param numpy.ndarray x_bins: The x-coordinates of the boundaries of the location bins.
    :param numpy.ndarray y_bins: The y-coordinates of the boundaries of the location bins.
    :param numpy.ndarray time_bins: The endpoints of the time bins.
    :param str stage: The stage in the SLiM life cycle that the endpoints of
        the time bins refer to (either "early" or "late"; defaults to "late").
    :param str remembered_stage: The stage in the SLiM life cycle during which
        individuals were Remembered (defaults to the stage the tree sequence was
        recorded at, stored in metadata).
    '''

    # TODO: make a helper function to do this
    if stage not in ("late", "early"):
        raise ValueError(f"Unknown stage '{stage}': "
                          "should be either 'early' or 'late'.")

    if remembered_stage is None:
        remembered_stage = ts.metadata['SLiM']['stage']

    if remembered_stage not in ("late", "early"):
        raise ValueError(f"Unknown remembered_stage '{remembered_stage}': "
                          "should be either 'early' or 'late'.")
    if remembered_stage != ts.metadata['SLiM']['stage']:
        warnings.warn(f"Provided remembered_stage '{remembered_stage}' does not"
                      " match the stage at which the tree sequence was saved"
                      f" ('{ts.metadata['SLiM']['stage']}'). This is not necessarily"
                      " an error, but mismatched stages will lead to inconsistencies:"
                      " make sure you know what you're doing.")

    # see individuals_alive_at for explanation
    if stage == "early" and ts.metadata['SLiM']['model_type'] == "WF":
        birth_offset = 1
    else:
        birth_offset = 0
    birth_times = ts.individuals_time - birth_offset
    if (ts.metadata['SLiM']['model_type'] == "WF"
            or stage == remembered_stage):
        age_offset = 0
    else:
        if (remembered_stage == "early"
                and stage == "late"):
            age_offset = -1
        else:
            age_offset = 1
    ages = individual_ages(ts) + age_offset
    death_times = birth_times - ages

    time_breaks = time_bins
    x_breaks = x_bins
    y_breaks = y_bins

    nxbins = len(x_breaks) - 1
    nybins = len(y_breaks) - 1
    ntbins = len(time_breaks) - 1
    popsize = np.empty((nxbins, nybins, ntbins))

    locations = ts.individuals_location

    for i in np.arange(nxbins):
        for j in np.arange(nybins):
            x0, x1 = x_breaks[i], x_breaks[i + 1]
            y0, y1 = y_breaks[j], y_breaks[j + 1]
            in_bin = _in_location_bin(locations, x0, x1, y0, y1)
            for k in np.arange(ntbins):
                t0, t1 = time_breaks[k], time_breaks[k + 1]
                popsize[i, j, k] = _average_time_alive(
                        birth_times[in_bin],
                        death_times[in_bin],
                        t0,
                        t1,
                )

    return popsize



"""
Test cases for the metadata reading/writing of pyslim.
"""
import json
import numpy as np
import pytest
import random
import warnings

import msprime
import tskit

import pyslim
import tests
from .recipe_specs import restarted_recipe_eq

class TestAnnotate(tests.PyslimTestCase):
    '''
    Tests for tools to annotate existing msprime-derived tree sequences.
    '''

    def verify_annotated_tables(self, ts1, ts2, check_alleles=True):
        '''
        Verify that the tables returned after annotation are equal, up to the
        expected forgetting of metadata, flags, and other things.
        '''
        assert pyslim.is_current_version(ts2)
        tables1 = ts1.dump_tables()
        tables2 = ts2.dump_tables()
        # compare nodes
        assert np.array_equal(tables1.nodes.flags, tables2.nodes.flags)
        assert np.array_equal(tables1.nodes.time, tables2.nodes.time)
        assert np.array_equal(tables1.nodes.population, tables2.nodes.population)
        # compare edges
        assert tables1.edges == tables2.edges
        # compare sites
        assert np.array_equal(tables1.sites.position, tables2.sites.position)
        if check_alleles:
            assert np.array_equal(tables1.sites.ancestral_state, tables2.sites.ancestral_state)
            assert np.array_equal(tables1.sites.ancestral_state_offset,
                                  tables2.sites.ancestral_state_offset)
        # compare mutations
        assert np.array_equal(tables1.mutations.site, tables2.mutations.site)
        assert np.array_equal(tables1.mutations.node, tables2.mutations.node)
        assert np.array_equal(tables1.mutations.time, tables2.mutations.time)
        if check_alleles:
            assert np.array_equal(tables1.mutations.derived_state, tables2.mutations.derived_state)
            assert np.array_equal(tables1.mutations.derived_state_offset,
                                  tables2.mutations.derived_state_offset)

    def verify_annotated_trees(self, ts1, ts2):
        '''
        Verify the *trees* returned before and after annotation are equal.
        '''
        assert ts1.num_trees == ts2.num_trees
        for t1, t2 in zip(ts1.trees(), ts2.trees()):
            assert t1.length == t2.length
            assert t1.get_parent_dict() == t2.get_parent_dict()
            assert t1.total_branch_length == t2.total_branch_length

    def verify_defaults(self, ts):
        '''
        Verify the default values have been entered into metadata.
        '''
        do_pops = [False for _ in ts.populations()]
        for m in ts.mutations():
            md = m.metadata
            assert isinstance(md['mutation_list'], list)
            for mdl in md['mutation_list']:
                assert mdl["mutation_type"] == 0
                assert mdl["selection_coeff"] == 0.0
                assert mdl["subpopulation"] == tskit.NULL
        for n in ts.nodes():
            md = n.metadata
            if not n.is_sample():
                assert md is None
            else:
                assert md["is_null"] is False
                assert md["genome_type"] == pyslim.GENOME_TYPE_AUTOSOME
                do_pops[n.population] = True
        for ind in ts.individuals():
            md = ind.metadata
            assert np.array_equal(ind.location, [0, 0, 0])
            assert ind.flags == pyslim.INDIVIDUAL_ALIVE
            assert md["sex"] == pyslim.INDIVIDUAL_TYPE_HERMAPHRODITE
            assert md["flags"] == 0
            assert md["pedigree_p1"] == tskit.NULL
            assert md["pedigree_p2"] == tskit.NULL
        for pop in ts.populations():
            if do_pops[pop.id]:
                md = pop.metadata
                assert md["selfing_fraction"] == 0.0
                assert md["female_cloning_fraction"] == 0.0
                assert md["male_cloning_fraction"] == 0.0
                assert md["sex_ratio"] == 0.0
                assert md["bounds_x0"] == 0.0
                assert md["bounds_x1"] == 1.0
                assert md["bounds_y0"] == 0.0
                assert md["bounds_y1"] == 1.0
                assert md["bounds_z0"] == 0.0
                assert md["bounds_z1"] == 1.0
                assert len(md["migration_records"]) == 0

    def verify_provenance(self, ts):
        for u in ts.provenances():
            tskit.validate_provenance(json.loads(u.record))

    def verify_slim_restart_equality(self, in_ts, out_ts):
        """
        Check for equality, in everything but the last provenance.
        """
        assert in_ts.num_provenances + 1 == out_ts.num_provenances
        in_tables = in_ts.dump_tables()
        in_tables.sort()
        out_tables = out_ts.dump_tables()
        out_tables.sort()
        self.assertTableCollectionsEqual(in_tables, out_tables, skip_provenance=-1)

    def verify_remapping(self, ts, rts, subpop_map):
        # below we assume the restart has not done additional simulation
        do_remap = (subpop_map is not None)
        if not do_remap:
            subpop_map = { f"p{k}" : k for k in range(ts.num_populations) }
        fwd_map = {-1 : -1}
        rev_map = {-1 : -1}
        for pk in subpop_map:
            # keys are requried to be of the form "pK"
            rts_k = int(pk[1:])
            assert rts_k < rts.num_populations
            ts_k = subpop_map[pk]
            assert ts_k < ts.num_populations
            assert ts_k not in fwd_map
            fwd_map[ts_k] = rts_k
            assert rts_k not in rev_map
            rev_map[rts_k] = ts_k

        # all pops in ts should be present UNLESS they have no metadata
        for pop in ts.populations():
            assert pop.id in fwd_map or pop.metadata is None

        # any extra pops in rts should have empty metadata
        for pop in rts.populations():
            if pop.id not in rev_map:
                assert pop.metadata is None

        for ts_k, rts_k in fwd_map.items():
            if ts_k != -1:
                ts_pop = ts.population(ts_k)
                rts_pop = rts.population(rts_k)
                ts_md = ts_pop.metadata
                rts_md = rts_pop.metadata
                if ts_md is None:
                    assert rts_md is None
                else:
                    if do_remap and (
                            ts_md['name'] == f"p{ts_pop.id}"
                            or ts_md['name'] == f"pop_{ts_pop.id}"
                    ):
                        ts_md['name'] = f"p{rts_pop.id}"
                    assert ts_md['name'] == rts_md['name']
                    if "slim_id" not in ts_md:
                        assert ts_md == rts_md
                    else:
                        assert ts_md["slim_id"] == ts_pop.id
                        assert rts_md["slim_id"] == rts_pop.id
                        k = "migration_records"
                        if k in ts_md:
                            assert k in rts_md or len(ts_md[k]) == 0
                        else:
                            assert k not in rts_md or len(rts_md[k]) == 0
                        if k in ts_md and k in rts_md:
                            assert len(ts_md[k]) == len(rts_md[k])
                            for x, y in zip(ts_md[k], rts_md[k]):
                                assert x['migration_rate'] == y['migration_rate']
                                assert fwd_map[x['source_subpop']] == y['source_subpop']

        # check other tables have been remapped
        # (uses the fact that no additional simulation has been done)
        assert ts.num_nodes == rts.num_nodes
        for n, rn in zip(ts.nodes(), rts.nodes()):
            n.population = fwd_map[n.population]
            # there can be rounding error due to time shift
            assert np.isclose(n.time, rn.time, atol=1e-16)
            n.time = rn.time
            assert n == rn

        assert ts.num_migrations == rts.num_migrations
        for m, rm in zip(ts.migrations(), rts.migrations()):
            m.source = fwd_map[m.source]
            m.dest = fwd_map[m.dest]
            assert m == rm

        assert ts.num_mutations == rts.num_mutations
        for m, rm in zip(ts.mutations(), rts.mutations()):
            md = m.metadata
            rmd = rm.metadata
            assert len(md['mutation_list']) == len(rmd['mutation_list'])
            for x, y in zip(md['mutation_list'], rmd['mutation_list']):
                x['subpopulation'] = fwd_map[x['subpopulation']]
                assert x == y

        assert ts.num_individuals == rts.num_individuals
        for i, ri in zip(ts.individuals(), rts.individuals()):
            md = i.metadata
            rmd = ri.metadata
            md['subpopulation'] = fwd_map[md['subpopulation']]
            assert md == rmd


    def test_annotate_errors(self, helper_functions):
        for ts in helper_functions.get_msprime_examples():
            with pytest.raises(ValueError):
                _ = pyslim.annotate(ts, model_type="WF", tick=0)
            with pytest.raises(ValueError):
                _ = pyslim.annotate(ts, model_type="WF", tick=4.4)
            with pytest.raises(ValueError):
                _ = pyslim.annotate(ts, model_type="foo", tick=4)
            with pytest.raises(ValueError):
                _ = pyslim.annotate(ts, model_type=[], tick=4)
        # no individuals
        ts = msprime.simulate(4)
        with pytest.raises(ValueError) as except_info:
            _ = pyslim.annotate(ts, model_type="WF", tick=1)
        assert "individuals" in str(except_info)
        # non-integer positions
        ts = msprime.sim_mutations(
                msprime.sim_ancestry(4, random_seed=8),
                rate=10,
                discrete_genome=False,
                random_seed=9,
        )
        with pytest.raises(ValueError) as except_info:
            _ = pyslim.annotate(ts, model_type="WF", tick=1)
        assert "not at integer" in str(except_info)

    def test_warns_overwriting_mutations(self, helper_functions):
        ts = msprime.sim_ancestry(
                4,
                population_size=10,
                sequence_length=10,
                recombination_rate=0.01,
                random_seed=100,
        )
        ts = msprime.sim_mutations(
                ts,
                rate=1,
                random_seed=12,
                model=msprime.SLiMMutationModel(type=1)
        )
        assert ts.num_mutations > 0
        with pytest.warns(Warning, match="already has.*metadata"):
            slim_ts = pyslim.annotate(ts, model_type="WF", tick=1) 

    def test_just_simulate(self, helper_functions, tmp_path):
        ts = msprime.sim_ancestry(
                4,
                population_size=10,
                sequence_length=10,
                recombination_rate=0.01,
                random_seed=100,
        )
        ts = msprime.sim_mutations(ts, rate=0.1, random_seed=11)
        slim_ts = pyslim.annotate(ts, model_type="WF", tick=1) 
        loaded_ts = helper_functions.run_msprime_restart(slim_ts, tmp_path, WF=True)
        self.verify_annotated_trees(ts, loaded_ts)

    def test_basic_annotation(self, helper_functions, tmp_path):
        for ts in helper_functions.get_msprime_examples():
            for do_mutations in [False, True]:
                tick = 4
                cycle = 1
                stage = "late"
                slim_ts = pyslim.annotate(
                        ts, model_type="WF",
                        tick=tick,
                        cycle=cycle,
                        stage=stage,
                        annotate_mutations=do_mutations,
                )
                assert slim_ts.metadata['SLiM']['model_type'] == 'WF'
                assert slim_ts.metadata['SLiM']['tick'] == tick
                assert slim_ts.metadata['SLiM']['cycle'] == cycle
                assert slim_ts.metadata['SLiM']['stage'] == stage
                assert slim_ts.metadata['SLiM']['file_version'] == pyslim.slim_file_version
                self.verify_annotated_tables(ts, slim_ts, check_alleles=(not do_mutations))
                self.verify_annotated_trees(ts, slim_ts)
                if not do_mutations:
                    self.verify_haplotype_equality(ts, slim_ts)
                self.verify_defaults(slim_ts)
                self.verify_provenance(slim_ts)
                # try loading this into SLiM
                loaded_ts = helper_functions.run_msprime_restart(slim_ts, tmp_path, WF=True)
                self.verify_annotated_tables(loaded_ts, slim_ts)
                self.verify_annotated_trees(loaded_ts, slim_ts)
                self.verify_haplotype_equality(loaded_ts, slim_ts)

    def test_annotate_refseq(self):
        ts = msprime.sim_ancestry(2, sequence_length=10, random_seed=77)
        refseq = "A" * int(ts.sequence_length)
        ats = pyslim.annotate(
                ts, model_type="nonWF", tick=5, reference_sequence=refseq
        )
        assert ats.reference_sequence.data == refseq

    def test_annotate_individuals(self, helper_functions, tmp_path):
        # test the workflow of annotating defaults, then assigning sexes randomly
        for ts in helper_functions.get_msprime_examples():
            slim_ts = pyslim.annotate(ts, model_type="nonWF", tick=1, stage="early")
            tables = slim_ts.dump_tables()
            top_md = tables.metadata
            top_md['SLiM']['separate_sexes'] = True
            tables.metadata = top_md
            metadata = [ind.metadata for ind in tables.individuals]
            sexes = [random.choice([pyslim.INDIVIDUAL_TYPE_FEMALE, pyslim.INDIVIDUAL_TYPE_MALE])
                     for _ in metadata]
            for j in range(len(metadata)):
                metadata[j]["sex"] = sexes[j]
            ims = tables.individuals.metadata_schema
            tables.individuals.packset_metadata(
                    [ims.validate_and_encode_row(r) for r in metadata]
            )
            pop_metadata = [p.metadata for p in tables.populations]
            for j, md in enumerate(pop_metadata):
                # nonWF models always have this
                md['sex_ratio'] = 0.0
            pms = tables.populations.metadata_schema
            tables.populations.packset_metadata(
                    [pms.validate_and_encode_row(r) for r in pop_metadata]
            )
            new_ts = tables.tree_sequence()
            for j, ind in enumerate(new_ts.individuals()):
                md = ind.metadata
                assert md["sex"] == sexes[j]
            self.verify_annotated_tables(new_ts, slim_ts)
            self.verify_annotated_trees(new_ts, slim_ts)
            self.verify_haplotype_equality(new_ts, slim_ts)
            # try loading this into SLiM
            loaded_ts = helper_functions.run_msprime_restart(new_ts, tmp_path, sex="A")
            self.verify_trees_equal(new_ts, loaded_ts)

    @pytest.mark.skip(reason="not working")
    def test_annotate_XY(self, helper_functions, tmp_path):
        # This is Not Right because as written the null chromosomes have history;
        # we need to *only* simulate the non-null chromosomes.
        random.seed(8)
        for ts in helper_functions.get_msprime_examples():
            for genome_type in ["X", "Y"]:
                slim_ts = pyslim.annotate_defaults(ts, model_type="nonWF", slim_generation=1)
                tables = slim_ts.dump_tables()
                top_md = tables.metadata
                top_md['SLiM']['separate_sexes'] = True
                tables.metadata = top_md
                metadata = [ind.metadata for ind in tables.individuals]
                sexes = [random.choice([pyslim.INDIVIDUAL_TYPE_FEMALE, pyslim.INDIVIDUAL_TYPE_MALE])
                         for _ in metadata]
                for j in range(len(metadata)):
                    metadata[j]["sex"] = sexes[j]
                ims = tables.individuals.metadata_schema
                tables.individuals.packset_metadata(
                        [ims.validate_and_encode_row(r) for r in metadata])
                node_metadata = [n.metadata for n in tables.nodes]
                node_is_null = [False for _ in range(ts.num_nodes)]
                for j in range(slim_ts.num_individuals):
                    nodes = slim_ts.individual(j).nodes
                    node_metadata[nodes[0]]["genome_type"] = pyslim.GENOME_TYPE_X
                    node_metadata[nodes[0]]["is_null"] = (genome_type != "X")
                    if sexes[j] == pyslim.INDIVIDUAL_TYPE_MALE:
                        node_metadata[nodes[1]]["genome_type"] = pyslim.GENOME_TYPE_Y
                        node_metadata[nodes[1]]["is_null"] = (genome_type != "Y")
                    else:
                        node_metadata[nodes[1]]["genome_type"] = pyslim.GENOME_TYPE_X
                        node_metadata[nodes[1]]["is_null"] = (genome_type != "X")
                    for n in nodes:
                        node_is_null[n] = node_metadata[n]["is_null"]
                nms = tables.nodes.metadata_schema
                tables.nodes.packset_metadata(
                        [nms.validate_and_encode_row(r) for r in node_metadata])
                # update populations
                pop_metadata = [p.metadata for p in tables.populations]
                for j, md in enumerate(pop_metadata):
                    # nonWF models always have this
                    md['sex_ratio'] = 0.0
                pms = tables.populations.metadata_schema
                tables.populations.packset_metadata(
                        [pms.validate_and_encode_row(r) for r in pop_metadata])
                new_ts = tables.tree_sequence()
                self.verify_annotated_tables(new_ts, slim_ts)
                self.verify_annotated_trees(new_ts, slim_ts)
                self.verify_haplotype_equality(new_ts, slim_ts)
                # try loading this into SLiM
                loaded_ts = helper_functions.run_msprime_restart(
                    new_ts, tmp_path, sex=genome_type)
                self.verify_trees_equal(new_ts, loaded_ts)
                # these are *not* equal but only due to re-ordering of nodes and individuals
                # ... and for some reason, .subset( ) or .simplify( ) do not produce equality
                # self.assertTableCollectionsEqual(new_ts, loaded_ts,
                #         skip_provenance=-1, reordered_individuals=True)

    def test_annotate_nodes(self, helper_functions):
        # test workflow of annotating defaults and then editing node metadata
        for ts in helper_functions.get_msprime_examples():
            slim_ts = pyslim.annotate(ts, model_type="nonWF", tick=1)
            tables = slim_ts.dump_tables()
            metadata = [n.metadata for n in tables.nodes]
            gtypes = [
                    random.choice([pyslim.GENOME_TYPE_X, pyslim.GENOME_TYPE_Y])
                    for _ in metadata
            ]
            for md, g in zip(metadata, gtypes):
                if md is not None:
                    md["genome_type"] = g
            nms = tables.nodes.metadata_schema
            tables.nodes.packset_metadata(
                    [nms.validate_and_encode_row(r) for r in metadata]
            )
            new_ts = tables.tree_sequence()
            for x, g in zip(new_ts.nodes(), gtypes):
                if x.metadata is not None:
                    assert x.metadata["genome_type"] == g
            # not testing SLiM because needs annotation of indivs to make sense

    def test_annotate_mutations(self, helper_functions):
        for ts in helper_functions.get_msprime_examples():
            slim_ts = pyslim.annotate(ts, model_type="nonWF", tick=1)
            tables = slim_ts.dump_tables()
            metadata = [m.metadata for m in tables.mutations]
            selcoefs = [random.uniform(0, 1) for _ in metadata]
            for j in range(len(metadata)):
                metadata[j]['mutation_list'][0]["selection_coeff"] = selcoefs[j]
            ms = tables.mutations.metadata_schema
            tables.mutations.packset_metadata(
                    [ms.validate_and_encode_row(r) for r in metadata]
            )
            new_ts = tables.tree_sequence()
            for j, x in enumerate(new_ts.mutations()):
                md = x.metadata
                assert np.isclose(md['mutation_list'][0]["selection_coeff"], selcoefs[j])

    def test_dont_annotate_mutations(self):
        # Test the option to not overwrite mutation annotations
        ts = msprime.sim_ancestry(10, random_seed=5)
        ts = msprime.sim_mutations(ts, rate=5, random_seed=3)
        assert ts.num_mutations > 0
        tables = ts.dump_tables()
        pre_mutations = tables.mutations.copy()
        pyslim.annotate_tables(
                tables,
                model_type="WF",
                tick=1,
                annotate_mutations=False
        )
        # this is necessary because b'' actually is decoded to
        # an empty mutation_list by the schema
        pre_mutations.metadata_schema = tables.mutations.metadata_schema
        assert tables.mutations.equals(pre_mutations)

    def test_annotate_empty(self):
        t = tskit.TableCollection(sequence_length=10)
        at = pyslim.annotate_tables(t, model_type='nonWF', tick=1)
        assert t.metadata_schema == pyslim.slim_metadata_schemas['tree_sequence']
        for x in ("population", "individual", "node", "site", "mutation"):
            assert getattr(t, x + "s").metadata_schema == pyslim.slim_metadata_schemas[x]

    def test_no_populations_errors(self, helper_functions, tmp_path):
        # test SLiM errors when individuals are in non-SLiM populations
        ts = msprime.sim_ancestry(5, population_size=10, sequence_length=100, random_seed=456)
        t = ts.dump_tables()
        p = t.populations.add_row(metadata={'name': '', 'description': ''})
        i = t.individuals.add_row()
        t.nodes.add_row(time=0.0, flags=tskit.NODE_IS_SAMPLE, individual=i)
        t.nodes.add_row(time=0.0, flags=tskit.NODE_IS_SAMPLE, individual=i)
        pyslim.annotate_tables(t, model_type='nonWF', tick=1, stage="late")
        pop = t.populations[-1]
        t.populations[-1] = pop.replace(metadata = {'name' : 'not_slim', 'description' : ''})
        ind = t.individuals[-1]
        md = ind.metadata
        md['subpopulation'] = (t.populations.num_rows - 1)
        t.individuals[-1] = ind.replace(metadata=md)
        ts = t.tree_sequence()
        with pytest.raises(RuntimeError, match='individual has a subpopulation id .1.'):
            _ = helper_functions.run_slim_restart(
                    ts,
                    "restart_nonWF.slim",
                    tmp_path,
                    WF=False,
            )

    def test_empty_populations_errors(self, helper_functions, tmp_path):
        # test SLiM errors on having empty populations in a WF model
        ts = msprime.sim_ancestry(5, population_size=10, sequence_length=100, random_seed=455)
        ts = pyslim.annotate(ts, model_type='WF', tick=1, stage="late")
        # first make sure does not error without the empty pops
        rts = helper_functions.run_slim_restart(
                ts,
                "restart_WF.slim",
                tmp_path,
                WF=True,
        )
        assert rts.num_populations == 1
        # now add empty subpops
        t = ts.dump_tables()
        for k in range(5):
            md = pyslim.default_slim_metadata('population')
            md['slim_id'] = k + 1
            md['name'] = f"new_pop_num_{k}"
            t.populations.add_row(metadata=md)
        ts = t.tree_sequence()
        for p in ts.populations():
            assert "slim_id" in p.metadata
        with pytest.raises(RuntimeError, match='subpopulation.*empty'):
            _ = helper_functions.run_slim_restart(
                    ts,
                    "restart_WF.slim",
                    tmp_path,
                    WF=True,
            )

    def test_empty_populations(self, helper_functions, tmp_path):
        # test SLiM doesn't error on having empty populations in a nonWF model
        # however, it may rewrite the metadata
        ts = msprime.sim_ancestry(5, population_size=10, sequence_length=100, random_seed=456)
        ts = pyslim.annotate(ts, model_type='nonWF', tick=1)
        t = ts.dump_tables()
        for k in range(5):
            md = pyslim.default_slim_metadata('population')
            md['slim_id'] = k + 1
            md['name'] = f"new_pop_num_{k}"
            md['description'] = f"the {k}-th added pop"
            t.populations.add_row(metadata=md)
        ts = t.tree_sequence()
        for subpop_map in (
                None,
                {"p0": 0, "p1": 1, "p2": 2, "p3": 3, "p4": 4, "p5": 5},
                {"p0": 0, "p2": 1, "p1": 2, "p9": 3, "p4": 4, "p6": 5},
                {"p10": 0, "p2": 1, "p1": 2, "p9": 3, "p4": 4, "p6": 5},
            ):
            sts = helper_functions.run_slim_restart(
                    ts,
                    "restart_nonWF.slim",
                    tmp_path,
                    WF=False,
                    subpop_map=subpop_map,
            )
            self.verify_remapping(ts, sts, subpop_map)

    def test_many_populations(self, helper_functions, tmp_path):
        # test we can add more than one population and that names are preserved
        ts = msprime.sim_ancestry(5, population_size=10, sequence_length=100, random_seed=455)
        t = ts.dump_tables()
        for k in range(1, 6):
            md = pyslim.default_slim_metadata('population')
            md['name'] = f"new_pop_num_{k}"
            md['description'] = f"the {k}-th added pop"
            md['slim_id'] = k
            t.populations.add_row(metadata=md)
            i = t.individuals.add_row()
            for _ in range(2):
                t.nodes.add_row(flags=1, time=0.0, individual=i, population=k)
        ts = t.tree_sequence()
        ts = pyslim.annotate(ts, model_type='WF', tick=1)
        for ind in ts.individuals():
            assert ind.flags == pyslim.INDIVIDUAL_ALIVE
        sts = helper_functions.run_slim_restart(
                ts,
                "restart_WF.slim",
                tmp_path,
                WF=True,
        )
        for k in range(1, 6):
            md = sts.population(k).metadata
            assert md['name'] == f"new_pop_num_{k}"

    def test_many_species_errors(self, helper_functions, tmp_path):
        # should error when there are population conflicts between species
        fox_ts = msprime.sim_ancestry(5, population_size=10, sequence_length=100, random_seed=455)
        fox_ts = pyslim.annotate(fox_ts, model_type='WF', tick=1)
        mouse_ts = msprime.sim_ancestry(5, population_size=20, sequence_length=1000, random_seed=234)
        mouse_ts = pyslim.annotate(mouse_ts, model_type='WF', tick=1)
        with pytest.raises(RuntimeError, match='p0 has been used already'):
            _ = helper_functions.run_multispecies_slim_restart(
                    {'fox': fox_ts, 'mouse': mouse_ts},
                    'restart_multispecies_WF.slim',
                    tmp_path,
                    WF=True,
            )
        with pytest.raises(RuntimeError, match='p1 has been used already'):
            _ = helper_functions.run_multispecies_slim_restart(
                    {'fox': fox_ts, 'mouse': mouse_ts},
                    'restart_multispecies_WF.slim',
                    tmp_path,
                    WF=True,
                    subpop_map={'fox': {'p1': 0}, 'mouse': {'p1': 0}},
            )

    def test_many_species(self, helper_functions, tmp_path):
        # verify we can load more than one species
        for wf in (True, False):
            fox_ts = msprime.sim_ancestry(5, population_size=10, sequence_length=100, random_seed=455)
            mouse_ts = msprime.sim_ancestry(5, population_size=20, sequence_length=1000, random_seed=234)
            fox_ts = pyslim.annotate(fox_ts, model_type="WF" if wf else "nonWF", tick=1)
            mouse_ts = pyslim.annotate(mouse_ts, model_type="WF" if wf else "nonWF", tick=1)
            for subpop_map in (
                        {'fox' : None, 'mouse' : {'p1' : 0}},
                        {'fox' : {'p1' : 0}, 'mouse' : None},
                        {'fox' : {'p3' : 0}, 'mouse' : {'p0' : 0}},
                    ):
                tsd = helper_functions.run_multispecies_slim_restart(
                        {'fox': fox_ts, 'mouse': mouse_ts},
                        'restart_multispecies_WF.slim' if wf else 'restart_multispecies_nonWF.slim',
                        tmp_path,
                        WF=wf,
                        subpop_map = subpop_map,
                )
                self.verify_remapping(fox_ts, tsd['fox'], subpop_map['fox'])
                self.verify_remapping(mouse_ts, tsd['mouse'], subpop_map['mouse'])

    def test_keeps_extra_populations(self, helper_functions, tmp_path):
        # test that non-SLiM metadata is not removed
        d = msprime.Demography()
        d.add_population(initial_size=10, name="A", extra_metadata={"pi": 3.1415})
        md = pyslim.default_slim_metadata('population')
        del md['name']
        del md['description']
        md['slim_id'] = 1
        d.add_population(initial_size=10, name="B", extra_metadata=md)
        d.add_population(initial_size=10, name="C", extra_metadata={"abc": 123})
        ts = msprime.sim_ancestry(
                samples={"B": 5},
                demography=d,
                sequence_length=100,
                random_seed=455
        )
        ts = pyslim.annotate(ts, model_type='WF', tick=1)
        sts = helper_functions.run_slim_restart(
                ts,
                "restart_WF.slim",
                tmp_path,
                WF=True,
        )
        assert sts.num_populations == 3
        assert np.all(ts.tables.nodes.population == 1)
        p = sts.population(0)
        assert p.metadata['name'] == "A"
        assert p.metadata['pi'] == 3.1415
        p = sts.population(1)
        assert p.metadata['name'] == "B"
        p = sts.population(2)
        assert p.metadata['name'] == "C"
        assert p.metadata['abc'] == 123

    def test_remapping_errors(self, helper_functions, tmp_path):
        d = msprime.Demography()
        d.add_population(initial_size=10, name="p0")
        d.add_population(initial_size=10, name="p1")
        d.add_population(initial_size=10, name="A")
        d.add_population_split(time=1.0, derived=["p0", "p1"], ancestral="A")
        ts = msprime.sim_ancestry(
                {"p0": 5, "p1": 5},
                demography=d,
                sequence_length=100,
                random_seed=455,
        )
        ts = pyslim.annotate(ts, model_type='nonWF', tick=1, stage="late")
        # test must refer to all subpopulations
        with pytest.raises(RuntimeError, match='subpopulation id 2 is used.*but is not remapped'):
            _ = helper_functions.run_slim_restart(
                    ts,
                    "restart_nonWF.slim",
                    tmp_path,
                    WF=False,
                    subpop_map = {
                        "p0" : 0,
                        "p1" : 1,
                    }
            )
        # test can't map the same subpop to two different pops
        with pytest.raises(RuntimeError, match='subpopMap value .0. is not unique'):
            _ = helper_functions.run_slim_restart(
                    ts,
                    "restart_nonWF.slim",
                    tmp_path,
                    WF=False,
                    subpop_map = {
                        "p0" : 0,
                        "p1" : 0,
                        "p2" : 1,
                        "p3" : 2,
                    }
            )
        # test errors if it refers to non-existing subpop
        with pytest.raises(RuntimeError):
            _ = helper_functions.run_slim_restart(
                    ts,
                    "restart_nonWF.slim",
                    tmp_path,
                    WF=False,
                    subpop_map = {
                        "p0" : 0,
                        "p1" : 1,
                        "p2" : 2,
                        "p3" : 3,
                    }
            )

    def test_remapping_empty_pops(self, helper_functions, tmp_path):
        # test that subpop_map doesn't need to explicitly refer to empty pops
        d = msprime.Demography()
        d.add_population(initial_size=10, name="p0")
        d.add_population(initial_size=10, name="p1")
        d.add_population(initial_size=10, name="p2") # unused
        d.add_population(initial_size=10, name="A")
        d.add_population_split(time=1.0, derived=["p0", "p1"], ancestral="A")
        ts = msprime.sim_ancestry(
                {"p0": 5, "p1": 5},
                demography=d,
                sequence_length=100,
                random_seed=456,
        )
        ts = pyslim.annotate(ts, model_type='nonWF', tick=1, stage="late")
        t = ts.dump_tables()
        r = t.populations[2].replace(metadata=None)
        # make both pop 2 and 4 empty
        t.populations[2] = r
        t.populations.append(r)
        ts = t.tree_sequence()
        for subpop_map in (
                {"p0" : 0, "p1" : 1, "p2" : 2, "p3" : 3, "p4" : 4},
                    {"p0" : 0, "p1" : 1, "p3": 3},
                ):
            rts = helper_functions.run_slim_restart(
                    ts,
                    "restart_nonWF.slim",
                    tmp_path,
                    WF=False,
                    subpop_map = subpop_map
            )
            self.verify_remapping(ts, rts, subpop_map)

    def test_remapping(self, helper_functions, tmp_path):
        d = msprime.Demography()
        # pop 0
        d.add_population(initial_size=10, name="A", extra_metadata={"pi": 3.1415})
        # pop 1
        md = pyslim.default_slim_metadata('population')
        del md['name']
        del md['description']
        md['slim_id'] = 1
        md['migration_records'] = [
            {
                "migration_rate" : 0.1,
                "source_subpop" : 3,
            }
        ]
        d.add_population(initial_size=10, name="pop_1", extra_metadata=md)
        # pop 2
        d.add_population(initial_size=10, name="pop_2", extra_metadata={"abc": 123})
        # pop 3
        md = pyslim.default_slim_metadata('population')
        del md['name']
        del md['description']
        md['slim_id'] = 3
        md['migration_records'] = [
            {
                "migration_rate" : 0.9,
                "source_subpop" : 1,
            }
        ]
        d.add_population(initial_size=10, name="D", extra_metadata=md)
        # pop 4
        d.add_population(initial_size=10, name="XYZ", extra_metadata={"abc": 5123})
        d.add_population_split(time=1.0, derived=["pop_1", "D"], ancestral="A")
        ts = msprime.sim_ancestry(
                samples={"pop_1": 5, "D": 5},
                demography=d,
                sequence_length=100,
                random_seed=455
        )
        ts = pyslim.annotate(ts, model_type='WF', tick=1)
        ts = msprime.sim_mutations(
                ts,
                rate=1e-2,
                random_seed=9,
                model=msprime.SLiMMutationModel(type=1),
        )
        assert ts.num_mutations > 0
        for subpop_map in (
                None,
                {"p0" : 0, "p1" : 1, "p2" : 2, "p3" : 3, "p4" : 4},
                {"p0" : 1, "p1" : 0, "p2" : 2, "p3" : 3, "p4" : 4},
                {"p5" : 0, "p0" : 1, "p7" : 2, "p2" : 3, "p9" : 4},
        ):
            rts = helper_functions.run_slim_restart(
                    ts,
                    "restart_WF.slim",
                    tmp_path,
                    WF=True,
                    subpop_map = subpop_map,
            )
            self.verify_remapping(ts, rts, subpop_map)

    @pytest.mark.skip(reason="migration table not supported by simplify")
    def test_remaps_migration_table(self, helper_functions, tmp_path):
        d = msprime.Demography()
        d.add_population(initial_size=10, name="p0")
        d.add_population(initial_size=10, name="p1")
        d.add_population(initial_size=10, name="A")
        d.add_population_split(time=1.0, derived=["p0", "p1"], ancestral="A")
        d.set_migration_rate(0, 1, 0.1)
        d.set_migration_rate(1, 0, 0.1)
        ts = msprime.sim_ancestry(
                {"p0": 5, "p1": 5},
                demography=d,
                sequence_length=100,
                random_seed=455,
                record_migrations=True
        )
        ts = pyslim.annotate(ts, model_type='WF', tick=1, stage="late")
        assert ts.num_migrations > 0
        rts = helper_functions.run_slim_restart(
                ts,
                "restart_WF.slim",
                tmp_path,
                WF=True,
                subpop_map = {"p5" : 0, "p2" : 1, "p0" : 2 },
        )
        self.verify_remapping(ts, rts, subpop_map)

    @pytest.mark.parametrize(
        'restart_name, recipe', restarted_recipe_eq("remove_subpop"), indirect=["recipe"])
    def test_keep_slim_names(self, restart_name, recipe, helper_functions, tmp_path):
        # test that SLiM retains names of SLiM populations
        # even when those populations are removed
        in_ts = recipe["ts"]
        # this recipe moves everyone to subpop 0
        out_ts = helper_functions.run_slim_restart(in_ts, restart_name, tmp_path)
        for n in out_ts.samples(time=0):
            assert out_ts.node(n).population == 0
        for j in range(1, in_ts.num_populations):
            in_md = in_ts.population(j).metadata
            out_md = out_ts.population(j).metadata
            assert in_md['name'] == out_md['name']

    @pytest.mark.parametrize(
        'restart_name, recipe', restarted_recipe_eq("no_op"), indirect=["recipe"])
    def test_reload_recapitate(
        self, restart_name, recipe, helper_functions, tmp_path
    ):
        # Test the ability of SLiM to load our files after recapitation.
        ts = recipe["ts"]
        # recapitate, reload
        in_ts = self.do_recapitate(ts, recombination_rate=1e-2, ancestral_Ne=10, random_seed=25)
        # put it through SLiM (which just reads in and writes out)
        print("IN")
        for p in in_ts.populations():
            print(p)
        out_ts = helper_functions.run_slim_restart(in_ts, restart_name, tmp_path)
        # check for equality, in everything but the last provenance
        self.verify_slim_restart_equality(in_ts, out_ts)
        # test for restarting with a subpop map
        for j, k in [(1, 7), (5, 7)]:
            subpop_map = {
                    f"p{i * j % k}" : i
                    for i in range(in_ts.num_populations)
            }
            out_ts = helper_functions.run_slim_restart(
                    in_ts,
                    restart_name,
                    tmp_path,
                    subpop_map=subpop_map
            )
            self.verify_remapping(in_ts, out_ts, subpop_map)

    @pytest.mark.parametrize(
        'restart_name, recipe', restarted_recipe_eq("no_op"), indirect=["recipe"])
    def test_reload_annotate(
        self, restart_name, recipe, helper_functions, tmp_path
    ):
        # Test the ability of SLiM to load our files after annotation.
        ts = recipe["ts"]
        tables = ts.dump_tables()
        metadata = [m.metadata for m in tables.mutations]
        has_nucleotides = tables.metadata['SLiM']['nucleotide_based']
        if has_nucleotides:
            nucs = [random.choice([0, 1, 2, 3]) for _ in metadata]
            refseq = "".join(
                    random.choices(
                        pyslim.NUCLEOTIDES,
                        k = int(ts.sequence_length),
                    ),
            )
            for n, md in zip(nucs, metadata):
                for m in md['mutation_list']:
                    m["nucleotide"] = n
            tables.reference_sequence.data = refseq
        for md in metadata:
            for m in md['mutation_list']:
                m["selection_coeff"] = random.random()
        ms = tables.mutations.metadata_schema
        tables.mutations.packset_metadata(
                [ms.validate_and_encode_row(r) for r in metadata]
        )
        in_ts = tables.tree_sequence()
        # put it through SLiM (which just reads in and writes out)
        out_ts = helper_functions.run_slim_restart(in_ts, restart_name, tmp_path)
        # check for equality, in everything but the last provenance
        self.verify_slim_restart_equality(in_ts, out_ts)


class TestReload(tests.PyslimTestCase):
    '''
    Tests for basic things related to reloading with SLiM
    '''

    @pytest.mark.parametrize(
        'restart_name, recipe', restarted_recipe_eq("no_op"), indirect=["recipe"])
    def test_load_without_provenance(
        self, restart_name, recipe, helper_functions, tmp_path
    ):
        in_ts = recipe["ts"]
        in_tables = in_ts.dump_tables()
        in_tables.provenances.clear()
        in_tables.sort()
        cleared_ts = in_tables.tree_sequence()
        out_ts = helper_functions.run_slim_restart(cleared_ts, restart_name, tmp_path)
        out_tables = out_ts.dump_tables()
        out_tables.provenances.clear()
        out_tables.sort()
        in_tables.assert_equals(out_tables)

    @pytest.mark.parametrize(
        'restart_name, recipe', restarted_recipe_eq("no_op", "nucleotides"), indirect=["recipe"])
    def test_reload_reference_sequence(
        self, restart_name, recipe, helper_functions, tmp_path
    ):
        in_ts = recipe["ts"]
        out_ts = helper_functions.run_slim_restart(in_ts, restart_name, tmp_path)
        assert in_ts.metadata['SLiM']['nucleotide_based'] is True
        assert out_ts.metadata['SLiM']['nucleotide_based'] is True
        assert in_ts.has_reference_sequence() == out_ts.has_reference_sequence()
        if in_ts.has_reference_sequence():
            in_ts.reference_sequence.assert_equals(out_ts.reference_sequence)

    @pytest.mark.parametrize(
        'restart_name, recipe', restarted_recipe_eq(), indirect=["recipe"])
    def test_restarts_and_runs_simplified(
        self, restart_name, recipe, helper_functions, tmp_path
    ):
        in_ts = recipe["ts"]
        n = set(in_ts.samples()[:4])
        for k in n:
            n |= set(in_ts.individual(in_ts.node(k).individual).nodes)
        py_ts = in_ts.simplify(list(n), filter_populations=False)
        out_ts = helper_functions.run_slim_restart(
                py_ts,
                restart_name,
                tmp_path,
        )
        assert out_ts.metadata["SLiM"]["tick"] >= in_ts.metadata["SLiM"]["tick"]




# Keywords: Python, tree-sequence recording, tree sequence recording

import msprime, pyslim

ts = msprime.sim_ancestry(samples=5000, population_size=5000, sequence_length=1e8,
    recombination_rate=1e-8)
slim_ts = pyslim.annotate(ts, model_type="WF", tick=1)
slim_ts.dump("recipe_17.8.trees")



# Keywords: Python, tree-sequence recording, tree sequence recording

# This is a Python recipe; note that it runs the SLiM model internally, below

import subprocess, tskit
import matplotlib.pyplot as plt
import numpy as np

# Run the SLiM model and load the resulting .trees file
#subprocess.check_output(["slim", "-m", "-s", "0", "./recipe_17.5.slim"])
ts = tskit.load("./recipe_17.5.trees")

# Load the .trees file and assess true local ancestry
breaks = np.zeros(ts.num_trees + 1)
ancestry = np.zeros(ts.num_trees + 1)
for tree in ts.trees():
    subpop_sum, subpop_weights = 0, 0
    for root in tree.roots:
        leaves_count = tree.num_samples(root) - 1  # subtract one for the root, which is a sample
        subpop_sum += tree.population(root) * leaves_count
        subpop_weights += leaves_count
    breaks[tree.index] = tree.interval[0]
    ancestry[tree.index] = subpop_sum / subpop_weights
breaks[-1] = ts.sequence_length
ancestry[-1] = ancestry[-2]

# Make a simple plot
plt.plot(breaks, ancestry)
# plt.show()
plt.savefig("./recipe_17.5.png")



import msprime
import tskit
import warnings
import numpy as np

from .slim_tree_sequence import *
from .slim_metadata import *
from .provenance import *
from .util import *


def recapitate(ts,
               ancestral_Ne=None,
               **kwargs
    ):
    '''
    Returns a "recapitated" tree sequence, by using msprime to run a
    coalescent simulation from the "top" of this tree sequence, i.e.,
    allowing any uncoalesced lineages to coalesce.

    To allow recapitation to be done correctly, the nodes of the
    first generation of the SLiM simulation from whom all samples inherit
    are still present in the tree sequence, but are not marked as samples.
    If you simplify the tree sequence before recapitating you must ensure
    these are not removed, which you do by passing the argument
    ``keep_input_roots=True`` to :meth:`simplify <tskit.TreeSequence.simplify>`.

    If you specify an ``ancestral_Ne``, then the recapitated portion of the
    tree sequence will be simulated in a single population with this
    (diploid) size. In other words, all lineages are moved to a single
    population of this size (named "ancestral" if this name is not already
    taken), and coalescence is allowed to happen.

    You may control the ancestral demography by passing in a ``demography``
    argument: see :func:`msprime.sim_ancestry`.
    
    In general, all defaults are whatever the defaults of
    {meth}`msprime.sim_ancestry` are; this includes recombination rate, so
    that if neither ``recombination_rate`` or a ``recombination_map`` are
    provided, there will be *no* recombination.

    :param tskit.TreeSequence ts: The tree sequence to transform.
    :param float ancestral_Ne: If specified, then will simulate from a single
        ancestral population of this size. It is an error to specify this
        as well as ``demography``.
    :param dict kwargs: Any other arguments to :func:`msprime.sim_ancestry`.
    '''
    is_current_version(ts, _warn=True)
    if ancestral_Ne is not None:
        if "demography" in kwargs:
            raise ValueError("You cannot specify both `demography` and `ancestral_Ne`.")
        recap_time = ts.metadata['SLiM']['tick']
        # In various circumstances depending on the stage in which the simulation was
        # started and when the tree sequence was written out, the time of the roots
        # might be one or even two less than the "tick" value. We have access
        # to the stage the tree sequence was written out, but not the time it
        # was *started*. So, we'll just check if we need to subtract one or two.
        # Consistency checking this requires looping over all the trees, unfortunately,
        # but it avoids some common errors.
        root_times = list(set([ts.node(n).time for t in ts.trees() for n in t.roots]))
        for adj in (1, 2):
            if np.allclose(root_times, recap_time - adj):
                recap_time -= adj
        if len(root_times) > 1 or not np.allclose(root_times, recap_time):
            message = (
                "Not all roots of the provided tree sequence are at the time expected "
                "by recapitate(). This could happen if you've simplified in "
                "python before recapitating (fix: don't simplify first). "
                "If could also happen in other situations, e.g., "
                "you added new individuals without parents in SLiM "
                "during the course of the simulation with sim.addSubPop(), "
                "in which case you will probably need to recapitate with "
                "msprime.sim_ancestry(initial_state=ts, ...). "
                f"(Expected root time: {recap_time}; "
                f"Observed root times: {root_times})"
            )
            raise ValueError(message)
        demography = msprime.Demography.from_tree_sequence(ts)
        # must set pop sizes to >0 even though we merge immediately
        for pop in demography.populations:
            pop.initial_size=1.0
        ancestral_name = "ancestral"
        derived_names = [pop.name for pop in demography.populations]
        while ancestral_name in derived_names:
            ancestral_name = (ancestral_name + "_ancestral")
        demography.add_population(
                name=ancestral_name,
                description="ancestral population simulated by msprime",
                initial_size=ancestral_Ne,
        )
        # the split has to come slightly longer ago than slim's tick
        # since that's when all the linages are at, and otherwise the event
        # won't apply to them
        demography.add_population_split(
                np.nextafter( recap_time, 2 * recap_time),
                derived=derived_names,
                ancestral=ancestral_name,
        )
        kwargs["demography"] = demography

    recap = msprime.sim_ancestry(
                initial_state = ts,
                **kwargs)

    return recap


def convert_alleles(ts):
    """
    Returns a modified tree sequence in which alleles have been replaced by
    their corresponding nucleotides. For sites, SLiM-produced tree sequences
    have "" (the empty string) for the ancestral state at each site; this method
    will replace this with the corresponding nucleotide from the reference sequence.
    For mutations, SLiM records the 'derived state' as a SLiM mutation ID; this
    method will this with the nucleotide from the mutation's metadata.

    This operation is not reversible: since SLiM mutation IDs are lost, the tree
    sequence will not be able to be read back into SLiM.

    The main purpose of this method is for output: for instance, this code will produce
    a VCF file with nucleotide alleles:

    .. code-block:: python

        nts = pyslim.convert_alleles(ts)
        with open('nucs.vcf', 'w') as f:
            nts.write_vcf(f)

    This method will produce an error if the tree sequence does not have a
    valid reference sequence or if any mutations do not have nucleotides: to first
    generate these, see :func:`.generate_nucleotides`.

    :param tskit.TreeSequence ts: The tree sequence to transform.
    """
    tables = ts.dump_tables()
    has_refseq = (
            ts.has_reference_sequence()
            and len(ts.reference_sequence.data) > 0
    )
    if not has_refseq:
        raise ValueError("Tree sequence must have a valid reference sequence.")
    # unfortunately, nucleotide mutations may be stacked (e.g., substitutions
    # will appear this way) and they don't appear in any particular order;
    # so we must guess which is the most recent, by choosing the one that
    # has the largest SLiM time, doesn't appear in the parent list, or has
    # the lagest SLiM ID.
    nuc_inds = tables.mutations.metadata_vector(['mutation_list', 0, 'nucleotide'], dtype='int')
    num_stacked = np.array([len(m.metadata['mutation_list']) for m in ts.mutations()])
    for k in np.where(num_stacked > 1)[0]:
        mut = ts.mutation(k)
        if mut.parent == tskit.NULL:
            pids = []
        else:
            pids = ts.mutation(mut.parent).derived_state.split(",")
        x = [
            (
                md['slim_time'],
                i not in pids,
                int(i),
                j
            ) for j, (i, md) in
            enumerate(
                zip(mut.derived_state.split(","), mut.metadata['mutation_list'])
            )
        ]
        x.sort()
        j = x[-1][3]
        nuc_inds[k] = mut.metadata['mutation_list'][j]['nucleotide']
    if np.any(nuc_inds == -1):
        raise ValueError("All mutations must be nucleotide mutations.")
    da = np.array(NUCLEOTIDES)[nuc_inds]
    tables.mutations.packset_derived_state(da)
    k = tables.sites.position.astype('int')
    aa = np.frombuffer(ts.reference_sequence.data.encode('utf-8'), dtype='S1')[k]
    tables.sites.packset_ancestral_state(aa.tobytes().decode('utf-8'))

    return tables.tree_sequence()


def generate_nucleotides(ts, reference_sequence=None, keep=True, seed=None):
    """
    Returns a modified tree sequence in which mutations have been randomly assigned nucleotides
    and (optionally) a reference sequence has been randomly generated.

    If ``reference_sequence`` is a string of nucleotides (A, C, G, and T) of
    length equal to the sequence length, this is used for the reference
    sequence. If no reference sequence is given, the reference_sequence property
    of ts is used if present; if not then a sequence of independent and
    uniformly random nucleotides is generated.

    SLiM stores the nucleotide as an integer in the mutation metadata, with -1 meaning "not
    a nucleotide mutation". This method assigns nucleotides by stepping through
    each mutation and picking a random nucleotide uniformly out of the three
    possible nucleotides that differ from the parental state (i.e., the derived
    state of the parental mutation, or the ancestral state if the mutation has
    no parent). If ``keep=True`` (the default), the mutations that already have a 
    nucleotide (i.e., an integer 0-3 in metadata) will not be modified.

    Technical note: in the case of stacked mutations, the SLiM mutation that
    determines the nucleotide state of the (tskit) mutation is the one with the largest
    slim_time attribute. This method tries to assign nucleotides so that each mutation
    differs from the previous state, but this is not always possible in certain
    unlikely cases.

    :param tskit.TreeSequence ts: The tree sequence to transform.
    :param bool reference_sequence: A reference sequence, or None to use an existing reference,
        or to randomly generate one.
    :param bool keep: Whether to leave existing nucleotides in mutations that already have one.
    :param int seed: The random seed for generating new alleles.
    """
    rng = np.random.default_rng(seed=seed)
    if reference_sequence is None:
        if not ts.has_reference_sequence():
            reference_sequence = rng.choice(
                    np.array([65, 67, 71, 84], dtype=np.int8),
                    int(ts.sequence_length),
                    replace=True,
            ).tobytes().decode('ascii')
    else:
        if len(reference_sequence) != ts.sequence_length:
            raise ValueError("Reference sequence must have length equal to sequence_length.")
        if len([x for x in reference_sequence if x not in NUCLEOTIDES]) > 0:
            raise ValueError("Reference sequence must be a string of A, C, G, and T only.")

    tables = ts.dump_tables()
    if reference_sequence is not None:
        tables.reference_sequence.data = reference_sequence
    tables.mutations.clear()
    sets = [[k for k in range(4) if k != i] for i in range(4)]
    states = np.full((ts.num_mutations,), -1)
    k = tables.sites.position.astype('int')
    aa_list = np.searchsorted(
            NUCLEOTIDES,
            np.frombuffer(tables.reference_sequence.data.encode('utf-8'), dtype='S1')[k],
    )
    for site in ts.sites():
        aa = aa_list[site.id]
        muts = {}
        for mut in site.mutations:
            if mut.parent == tskit.NULL:
                pa = aa
                pds = []
            else:
                pa = states[mut.parent]
                pds = ts.mutation(mut.parent).derived_state.split(",")
            this_da = pa
            ml = mut.metadata
            max_time = -np.inf
            for i, md in zip(mut.derived_state.split(","), ml['mutation_list']):
                da = md['nucleotide']
                if da == -1 or not keep:
                    if i in muts:
                        da = muts[i]
                    else:
                        da = sets[pa][rng.integers(3)]
                    md['nucleotide'] = da
                muts[i] = da
                # the official nucleotide state is from the SLiM mutation with
                # the largest slim_time attribute that was not present in the parent
                # mutation
                if md["slim_time"] >= max_time and i not in pds:
                    this_da = da
                    max_time = md["slim_time"]
            states[mut.id] = this_da
            tables.mutations.append(mut.replace(metadata=ml))

    return tables.tree_sequence()


def individual_ages(ts):
    """
    Returns the ages of all individuals in the tree sequence, extracted
    from metadata. The result is a array of length equal to the number of
    individuals, with k-th entry equal to ``ts.individual(k).metadata["age"]``.

    :return: An array of ages of individuals.
    """
    if ts.metadata['SLiM']['model_type'] != "WF":
        ages = ts.tables.individuals.metadata_vector("age")
    else:
        ages = np.zeros(ts.num_individuals, dtype='int')
    return ages


def individuals_alive_at(ts, time, stage='late', remembered_stage=None,
                         population=None, samples_only=False):
    """
    Returns an array giving the IDs of all individuals that are known to be
    alive at the given time ago.  This is determined using their birth time
    ago (given by their `time` attribute) and, for nonWF models,
    their `age` attribute (which is equal to their age at the last time
    they were Remembered). See also {func}`.individual_ages_at`.

    In WF models, birth occurs after "early()", so that individuals are only
    alive during "late()" for the time step when they have age zero,
    while in nonWF models, birth occurs before "early()", so they are alive
    for both stages.
    
    In both WF and nonWF models, mortality occurs between
    "early()" and "late()", so that individuals are last alive during the
    "early()" stage of the time step of their final age, and if individuals
    are alive during "late()" they will also be alive during "first()" and
    "early()" of the next time step. This means it is important to know during
    which stage individuals were Remembered - for instance, if the call to
    sim.treeSeqRememberIndividuals() was made during "early()" of a given time step,
    then those individuals might not have survived until "late()" of that
    time step. Since SLiM does not record the stage at which individuals
    were Remembered, you can specify this by setting ``remembered_stages``:
    it should be the stage during which *all* calls to
    ``sim.treeSeqRememberIndividuals()``,  as well as to ``sim.treeSeqOutput()``,
    were made.

    Note also that in nonWF models, birth occurs between "first()" and
    "early()", so the possible parents in a given time step are those that are
    alive in "early()" and have age greater than zero, or, equivalently, are
    alive in "late()" during the previous time step.
    In WF models, birth occurs after "early()", so possible parents in a
    given time step are those that are alive during "early()" of that time
    step or are alive during "late()" of the previous time step.

    Since individuals may be created not during the usual 'birth' stage
    by `addSubPop( )`, and the stage at which they are created is not stored,
    results may be unreliable for the first-generation individuals.

    :param tskit.TreeSequence ts: A tree sequence.
    :param float time: The number of ticks (i.e., time steps) ago.
    :param str stage: The stage in the SLiM life cycle that we are inquiring
        about (either "first", "early" or "late"; defaults to "late").
    :param str remembered_stage: The stage in the SLiM life cycle
        during which individuals were Remembered (defaults to the stage the
        tree sequence was recorded at, stored in metadata).
    :param int population: If given, return only individuals in the
        population(s) with these population ID(s).
    :param bool samples_only: Whether to return only individuals who have at
        least one node marked as samples.
    """
    is_current_version(ts, _warn=True)
    if stage not in ("late", "early", "first"):
        raise ValueError(f"Unknown stage '{stage}': "
                          "should be either 'first', 'early' or 'late'.")

    if remembered_stage is None:
        remembered_stage = ts.metadata['SLiM']['stage']

    if remembered_stage not in ("late", "early", "first"):
        raise ValueError(f"Unknown remembered_stage '{remembered_stage}': "
                          "should be either 'first', 'early' or 'late'.")
    if remembered_stage != ts.metadata['SLiM']['stage']:
        warnings.warn(f"Provided remembered_stage '{remembered_stage}' does not"
                      " match the stage at which the tree sequence was saved"
                      f" ('{ts.metadata['SLiM']['stage']}'). This is not necessarily"
                      " an error, but mismatched stages will lead to inconsistencies:"
                      " make sure you know what you're doing.")

    # An individual's tskit time is the tskit counter at the time of their birth.
    # The tskit counter clicks once at the start of each reproduction bout.
    # In a WF model, individuals are alive for all stages with the same value
    # of the tskit counter. In a nonWF model, an individual of age `a` will be
    # alive for `a + 1` 'early' events, and `a` 'late' and 'first' events,
    # starting with the one having the same value of the tskit counter.
    # Also note that if remembered_stage is 'late', then they are recorded
    # before their age counter clicks, so they will actually have age one greater.
    # Since `time` is in "number of time steps ago",
    # we are inquiring about SLiM tick `n - time`, where `n` is
    # the value of `ts.metadata["SLiM"]["tick"]`. To convert this along with
    # our stage information into a tskit time `t`,
    # let x = 1 if the stage is 'first' or (is 'early' and WF)
    # and y = 1 if remembered stage is 'late' or (is 'early' and nonWF);
    # then t = time + x + y - 1 .
    is_wf = (ts.metadata['SLiM']['model_type'] == "WF")
    x = (stage == "first" or (stage == "early" and is_wf))
    y = (remembered_stage == "late" or (remembered_stage == "early" and not is_wf))
    t = time + x + y - 1
    birth_times = ts.individuals_time
    ages = individual_ages(ts)
    if is_wf:
        alive_bool = (birth_times == t)
    else:
        age_offset = (stage == "early") + (remembered_stage == "late")
        alive_bool = np.logical_and(
                birth_times >= t,
                birth_times - ages < t + age_offset
        )
    if population is not None:
        alive_bool &= np.isin(
                ts.individuals_population,
                population
        )
    if samples_only:
        nodes = ts.tables.nodes
        alive_bool &= (
                0 < np.bincount(1 + nodes.individual,
                                nodes.flags & tskit.NODE_IS_SAMPLE,
                                minlength=1 + ts.num_individuals)[1:]
        )
    return np.where(alive_bool)[0]


def individual_ages_at(ts, time, stage="late", remembered_stage="late"):
    """
    Returns the `ages` of each individual at the corresponding time ago,
    which will be ``nan`` if the individual is either not born yet or dead.
    This is computed as the time ago the individual was born (found by the
    `time` associated with the the individual's nodes) minus the `time`
    argument; while "death" is inferred from the individual's ``age``,
    recorded in metadata. These values are the same as what would be shown
    in SLiM during the corresponding time step and stage.

    Since age increments at the end of each time step,
    the age is the number of time steps ends the individual has lived
    through, so if they were born in time step `time`, then their age
    will be zero.

    In a WF model, this method does not provide any more information than
    does {func}`.individuals_alive_at`, but for consistency, non-nan ages
    will be 0 in "late" and 1 in "first" and "early".
    See {func}`.individuals_alive_at` for further discussion.

    :param tskit.TreeSequence ts: A tree sequence.
    :param float time: The reference time ago.
    :param str stage: The stage in the SLiM life cycle used to determine who
        is alive (either "early" or "late"; defaults to "late").
    :param str remembered_stage: The stage in the SLiM life cycle during which
        individuals were Remembered.
    """
    ages = np.repeat(np.nan, ts.num_individuals)
    alive = individuals_alive_at(
            ts,
            time,
            stage=stage,
            remembered_stage=remembered_stage
    )
    # to convert individuals_time to number of ticks ago we subtract (y - 1), so
    is_wf = (ts.metadata['SLiM']['model_type'] == "WF")
    y = (remembered_stage == "late" or (remembered_stage == "early" and not is_wf))
    t = time + y - 1
    ages[alive] = ts.individuals_time[alive] - t
    return ages


def slim_time(ts, time, stage="late"):
    """
    Converts the given "tskit times" (i.e., in units of time before the end
    of the simulation) to SLiM times (those recorded by SLiM, usually in units
    of ticks since the start of the simulation). Although the latter are
    always integers, these will not be if the provided times are not integers.

    When the tree sequence is written out, SLiM records the current
    current tick in the metadata:
    ``ts.metadata['SLiM']['tick']``. In most cases, the "SLiM time"
    referred to by a time ago in the tree sequence (i.e., the value that would
    be reported by community.tick within SLiM at the point in time thus
    referenced) can be obtained by subtracting that time ago from
    ``ts.metadata['SLiM']['tick']``. However, in WF models, birth
    happens between the “early()” and “late()” stages, so if the tree
    sequence was written out using sim.treeSeqOutput() during “early()” in
    a WF model, the tree sequence’s times measure time before the last set
    of individuals are born, i.e., before SLiM time step
    ``ts.metadata['SLiM']['tick'] - 1``. The same thing applies to
    the "first" stage for both WF and nonWF models.

    In some situations (e.g., mutations added during early() in WF models)
    this may not return what you expect. See :ref:`sec_metadata_converting_times`
    for more discussion.

    :param tskit.TreeSequence ts: A SLiM-compatible TreeSequence.
    :param numpy.ndarray time: An array of times to be converted.
    :param str stage: The stage of the SLiM life cycle that the SLiM time
        should be computed for.
    """
    is_current_version(ts, _warn=True)
    is_wf = (ts.metadata['SLiM']['model_type'] == "WF")
    remembered_stage = ts.metadata['SLiM']['stage']
    x = (stage == "first" or (stage == "early" and is_wf))
    y = (remembered_stage == "late" or (remembered_stage == "early" and not is_wf))
    slim_time = ts.metadata['SLiM']['tick'] - time + x + y - 1
    return slim_time


def _do_individual_parents_stuff(ts, return_parents=False):
    # Helper for has_individual_parents and individual_parents,
    # which share a lot of machinery.
    tables = ts.tables
    edges = tables.edges
    nodes = tables.nodes
    edge_parent_indiv = nodes.individual[edges.parent]
    edge_child_indiv = nodes.individual[edges.child]
    # nodes whose parent nodes are all in the same individual
    unique_parent_nodes = unique_labels_by_group(
            edges.child,
            edge_parent_indiv,
            minlength=nodes.num_rows)
    unique_parent_edges = unique_parent_nodes[edges.child]
    # edges describing relationships between individuals
    indiv_edges = np.logical_and(
            np.logical_and(edge_parent_indiv != tskit.NULL,
                                 edge_child_indiv != tskit.NULL),
            unique_parent_edges)
    # individual edges where the parent was alive during "late"
    # of the time step before the child is born
    ind_times = ts.individuals_time
    ind_ages = individual_ages(ts)
    child_births = ind_times[edge_child_indiv[indiv_edges]]
    parent_births = ind_times[edge_parent_indiv[indiv_edges]]
    alive_edges = indiv_edges.copy()
    if ts.metadata['SLiM']['model_type'] == "WF":
        alive_edges[indiv_edges] = (child_births + 1 == parent_births)
    else:
        parent_deaths = parent_births - ind_ages[edge_parent_indiv[indiv_edges]]
        alive_edges[indiv_edges] = (child_births + 1 >= parent_deaths)
    edge_spans = edges.right - edges.left
    parental_span = np.bincount(edge_child_indiv[alive_edges],
            weights=edge_spans[alive_edges], minlength=ts.num_individuals)
    # we could also check for edges without individual parents terminating
    # in this individual, but this is unnecessary as the entire genome is
    # accounted for
    has_all_parents = (parental_span == 2 * ts.sequence_length)
    if return_parents:
        full_parent_edges = np.logical_and(
                alive_edges,
                has_all_parents[edge_child_indiv])
        parents = np.unique(np.column_stack(
                        [edge_parent_indiv[full_parent_edges],
                         edge_child_indiv[full_parent_edges]]
                        ), axis=0)
        return parents
    else:
        return has_all_parents


def individual_parents(ts):
    '''
    Finds all parent-child relationships in the tree sequence (as far as we
    can tell). The output will be a two-column array with row [i,j]
    indicating that individual i is a parent of individual j.  See
    {func}`.has_individual_parents` for exactly which parents are returned.

    See {func}`.individuals_alive_at` for further discussion about how
    this is determined based on when the individuals were Remembered.

    :param tskit.TreeSequence ts: A :class:`tskit.TreeSequence`.
    :return: An array of individual IDs, with row [i, j] if individual i is
        a parent of individual j.
    '''
    return _do_individual_parents_stuff(ts, return_parents=True)


def has_individual_parents(ts):
    '''
    Finds which individuals have both their parent individuals also present
    in the tree sequence, as far as we can tell. To do this, we return a
    boolean array with True for those individuals for which:

    - all edges terminating in that individual's nodes are in individuals,
    - each of the individual's nodes inherit from a single individual only,
    - those parental individuals were alive when the individual was born,
    - the parental individuals account for two whole genomes.

    This returns a boolean array indicating for each individual whether all
    these are true. Note in particular that individuals with only *one*
    recorded parent are *not* counted as "having parents".

    See {func}`.individuals_alive_at` for further discussion about how
    this is determined based on when the individuals were Remembered.

    :param tskit.TreeSequence ts: A :class:`tskit.TreeSequence`.
    :return: A boolean array of length equal to ``targets``.
    '''
    return _do_individual_parents_stuff(ts, return_parents=False)


def annotate(ts, **kwargs):
    '''
    Takes a tree sequence (as produced by msprime, for instance), and adds in the
    information necessary for SLiM to use it as an initial state, filling in
    mostly default values. Returns a :class:`tskit.TreeSequence`.

    :param tskit.TreeSequence ts: A :class:`tskit.TreeSequence`.
    :param str model_type: SLiM model type: either "WF" or "nonWF".
    :param int tick: What tick number in SLiM correponds to
        ``time=0`` in the tree sequence.
    :param int cycle: What cycle number in SLiM correponds to
        ``time=0`` in the tree sequence (default: equal to ``tick``).
    :param int stage: What stage in SLiM's cycle has the tree sequence
        been written out in (defaults to "early"; must be "early" or "late").
    :param str reference_sequence: A reference sequence of length
        equal to ts.sequence_length.
    :param bool annotate_mutations: Whether to replace mutation metadata
        with defaults. (If False, the mutation table is unchanged.)
    '''
    tables = ts.dump_tables()
    annotate_tables(tables, **kwargs)
    return tables.tree_sequence()


def annotate_tables(tables, model_type, tick, cycle=None, stage="early", reference_sequence=None,
        annotate_mutations=True):
    '''
    Does the work of :func:`annotate`, but modifies the tables in place: so,
    takes tables as produced by ``msprime``, and makes them look like the
    tables as output by SLiM. See :func:`annotate` for details.
    '''
    if stage not in ("early", "late"):
        raise ValueError(f"stage must be 'early' or 'late' (provided {stage})")
    if (type(tick) is not int) or (tick < 1):
        raise ValueError("tick must be an integer and at least 1.")
    # set_nodes must come before set_populations
    if model_type == "WF":
        default_ages = -1
    elif model_type == "nonWF":
        default_ages = 0
    else:
        raise ValueError("Model type must be 'WF' or 'nonWF'")
    if cycle is None:
        cycle = tick
    if not np.allclose(tables.sites.position, np.floor(tables.sites.position)):
        raise ValueError(
                "Site positions in this tree sequence are not at integer values, "
                "but must be for loading into SLiM: generate mutations with "
                "sim_mutations(..., discrete_genome=True), not simulate()."
        )
    top_metadata = default_slim_metadata('tree_sequence')['SLiM']
    top_metadata['model_type'] = model_type
    top_metadata['tick'] = tick
    top_metadata['cycle'] = cycle
    top_metadata['stage'] = stage
    set_tree_sequence_metadata(tables, **top_metadata)
    set_metadata_schemas(tables)
    _annotate_nodes_individuals(tables, age=default_ages)
    _annotate_populations(tables)
    if annotate_mutations:
        _annotate_sites_mutations(tables)
    if reference_sequence is not None:
        tables.reference_sequence.data = reference_sequence

def next_slim_mutation_id(ts):
    '''
    Returns the next SLiM mutation ID for this tree sequence. This is useful 
    because if you want to add more mutations to your SLiM tree sequence using 
    :func:`msprime.sim_mutations`, you may need to specify the parameter 
    `next_id` in your :class:`msprime.SLiMMutationModel` to be larger than any 
    existing mutation IDs. Setting `next_id` equal to the output of this 
    function will allow the mutated tree sequence to be read in by SLiM.
    To do this, recall that the "derived state" of SLiM's mutations are
    comma-separated strings of mutation IDs; this function just parses all derived
    states and returns one larger than the largest integer found. It will return an error
    if it encounters derived states that are not comma-separated strings of integers.
    '''
    max_id = 0
    for mut in ts.mutations():
        ds = mut.derived_state
        if len(ds) > 0:
            for d in ds.split(","):
                try:
                    max_id = max(max_id, int(d))
                except ValueError:
                    raise ValueError(f"The derived state of a mutation ({ds}) in the tree "
                                     "sequence is not a comma-separated list of values "
                                     "coercible to int. This is not a valid SLiM tree sequence.")
    return max_id + 1


def _annotate_nodes_individuals(tables, age):
    '''
    Adds to a TableCollection the information relevant to individuals required
    for SLiM to load in a tree sequence, that is found in Node and Individual
    tables.  This will replace the metadata in those tables. For this to work,
    assumes all sample nodes are in individuals.

    This is designed to make it easy to assign default values:
    - (location) individual locations to (0, 0, 0)
    - (age) individual age to 0
    - (ind_id) SLiM individual pedigree IDs to sequential integers starting from 0
    - (ind_population) individual populations to 0
    - (node_id) SLiM genome IDs to sequential integers starting with samples from 0
    - (node_is_null) genomes to be non-null
    - (node_type) genome type to 0 (= autosome)
    - (ind_flags) INDIVIDUAL_ALIVE

    If you have other situations, like non-alive "remembered" individuals, you
    will need to edit the tables by hand, afterwards.
    '''
    ind_population = np.full(tables.individuals.num_rows, -1, dtype="int")
    ind_slim_id = np.full(tables.individuals.num_rows, 0, dtype='int')
    nid = 0
    node_metadata = []
    for j, n in enumerate(tables.nodes):
        if n.flags & tskit.NODE_IS_SAMPLE > 0:
            i = n.individual
            if i == tskit.NULL:
                raise ValueError("To annotate, samples must be in individuals.")
            else:
                ind_population[i] = n.population
                ind_slim_id[i] = 1
            md = default_slim_metadata("node")
            md["slim_id"] = nid
            nid += 1
        else:
            md = n.metadata
        node_metadata.append(md)

    nms = tables.nodes.metadata_schema
    tables.nodes.packset_metadata([
        nms.validate_and_encode_row(x)
        for x in node_metadata
    ])

    slim_ind = (ind_slim_id != 0)
    ind_slim_id = np.cumsum(ind_slim_id) - 1

    ind_metadata = []
    ind_flags = tables.individuals.flags
    for j, ind in enumerate(tables.individuals):
        if slim_ind[j]:
            md = default_slim_metadata("individual")
            md["pedigree_id"] = int(ind_slim_id[j])
            md["subpopulation"] = int(ind_population[j])
            md["age"] = age
            # no packset analogue for flags, so do it here
            # and we don't have to resize for flags,
            # so no big deal
            ind_flags[j] |= INDIVIDUAL_ALIVE
        else:
            md = ind.metdadata
        ind_metadata.append(md)
    tables.individuals.set_columns(
        flags=ind_flags,
        parents=tables.individuals.parents,
        parents_offset=tables.individuals.parents_offset,
    )
    ims = tables.individuals.metadata_schema
    tables.individuals.packset_metadata([
            ims.validate_and_encode_row(x)
            for x in ind_metadata
    ])
    tables.individuals.packset_location(
            [[0.0] * 3 if si else [] for si in slim_ind]
    )


def _annotate_populations(tables):
    '''
    Adds to a TableCollection the information about populations required for SLiM
    to load a tree sequence. This will replace anything already in the Population
    table for populations referenced by nodes alive at time zero.
    '''
    alive_ind = (tables.individuals.flags & INDIVIDUAL_ALIVE > 0)
    do_pops = np.unique(
        tables.nodes.population[
            np.logical_and(
                tables.nodes.individual >= 0,
                alive_ind[tables.nodes.individual]
            )
        ]
    )
    for j, p in enumerate(tables.populations):
        if j in do_pops:
            if "slim_id" not in p.metadata:
                md = default_slim_metadata("population")
                md["slim_id"] = j
                md["name"] = f"pop_{j}"
                tables.populations[j] = p.replace(metadata=md)


def _annotate_sites_mutations(tables):
    '''
    Adds to a TableCollection the information relevant to mutations required
    for SLiM to load in a tree sequence. This means adding to the metadata column
    of the Mutation table,  It will also
    - give SLiM IDs to each mutation
    - replace ancestral states with ""
    This will replace any information already in the metadata or derived state
    columns of the Mutation table. We set slim_time in metadata so that
    - tick = floor(tskit time) + slim_time
    '''
    if len(tables.mutations.metadata) > 0:
        warnings.warn(
                "The provided tree sequence already has some mutations with "
                "metadata; this metadata will be overwritten."
        )
    num_mutations = tables.mutations.num_rows
    default_mut = default_slim_metadata("mutation_list_entry")
    dsb, dso = tskit.pack_bytes([str(j).encode() for j in range(num_mutations)])
    slim_time = tables.metadata["SLiM"]["tick"] - np.floor(tables.mutations.time).astype("int")
    mms = tables.mutations.metadata_schema
    mutation_metadata = [
            mms.encode_row(
                {"mutation_list": 
                 [{"mutation_type": default_mut["mutation_type"],
                  "selection_coeff": default_mut["selection_coeff"],
                  "subpopulation": default_mut["subpopulation"],
                  "slim_time": st,
                  "nucleotide": default_mut["nucleotide"]
                   }]
                })
            for st in slim_time]
    mdb, mdo = tskit.pack_bytes(mutation_metadata)
    tables.mutations.set_columns(
            site=tables.mutations.site,
            node=tables.mutations.node,
            time=tables.mutations.time,
            derived_state=dsb,
            derived_state_offset=dso,
            parent=tables.mutations.parent,
            metadata=mdb,
            metadata_offset=mdo)
    tables.sites.set_columns(
            position=tables.sites.position,
            ancestral_state=np.array([], dtype='int8'),
            ancestral_state_offset=np.zeros(tables.sites.num_rows + 1, dtype='uint32'))



# Keywords: Python, tree-sequence recording, tree sequence recording

import tskit

ts = tskit.load("recipe_17.8_II.trees").simplify()

for tree in ts.trees():
    for root in tree.roots:
        print(tree.time(root))



# Keywords: Python, nucleotide-based, nucleotide sequence, sequence-based mutation rate

import tskit, pyslim
import numpy as np

ts = tskit.load("recipe_nucleotides.trees")

M = [[0 for _ in pyslim.NUCLEOTIDES] for _ in pyslim.NUCLEOTIDES]
for mut in ts.mutations():
    mut_list = mut.metadata["mutation_list"]
    k = np.argmax([u["slim_time"] for u in mut_list])
    derived_nuc = mut_list[k]["nucleotide"]
    if mut.parent == -1:
        acgt = ts.reference_sequence.data[int(ts.site(mut.site).position)]
        parent_nuc = pyslim.NUCLEOTIDES.index(acgt)
    else:
        parent_mut = ts.mutation(mut.parent)
        assert(parent_mut.site == mut.site)
        parent_nuc = parent_mut.metadata["mutation_list"][0]["nucleotide"]
    M[parent_nuc][derived_nuc] += 1

print("{}\t{}\t{}".format('ancestral', 'derived', 'count'))
for j, a in enumerate(pyslim.NUCLEOTIDES):
    for k, b in enumerate(pyslim.NUCLEOTIDES):
        print("{}\t{}\t{}".format(a, b, M[j][k]))



"""
Test cases for tree sequences.
"""
import random
import os

import numpy as np
import pytest
import tskit
import msprime
import pyslim

import tests

from .recipe_specs import recipe_eq

class TestPopulationSize(tests.PyslimTestCase):

    def population_size_simple(self, ts, x_bins, y_bins, time_bins, **kwargs):
        '''
        Calculates population size in each location bin averaged over each time_bin.   
        '''
        
        # Want to return
        # [[[px0y0t0, px0y0t1, ...], [px0y1t0, px0y1t1, ...], ...],
        #  [[px1y0t0, px0y0t1, ...], [px1y1t0, px1y1t1, ...], ...],
        #  [[px2y0t0, px0y0t1, ...], [px2y1t0, px2y1t1, ...], ...]]
        # where pxiyjtk is the population size in [x_breaks[i], x_breaks[i + 1]) and [y_breaks[j], x_breaks[j + 1]),
        # averaged over each time point in [t[k], t[k + 1])
        # different offsets for different stages, from pyslim.individuals_alive_at code
               
        time_breaks = time_bins
        x_breaks = x_bins
        y_breaks = y_bins

        # Initialize array to store popoluationsize
        nxbins = len(x_breaks) - 1
        nybins = len(y_breaks) - 1
        ntbins = len(time_breaks) - 1
        popsize = np.empty((nxbins, nybins, ntbins))

        # Location, times, and ages of individuals
        locations = ts.individuals_location
        times = ts.individuals_times

        # Iterate through location bins and time bins
        for i in np.arange(nxbins):
            for j in np.arange(nybins):
                # Endpoints of bins
                x0, x1 = x_breaks[i], x_breaks[i + 1]
                y0, y1 = y_breaks[j], y_breaks[j + 1]
                for k in np.arange(ntbins):
                    # Endpoints of bins
                    t0, t1 = time_breaks[k], time_breaks[k + 1]
                    alive = 0
                    for t in np.arange(np.ceil(t0), t1):
                        for ind_id in pyslim.individuals_alive_at(ts, t, **kwargs):
                            ind = ts.individual(ind_id)
                            if (ind.location[0] < x1 and ind.location[0] >= x0 and ind.location[1] < y1 and ind.location[1] >= y0):
                                alive += 1/(t1-t0)
                    popsize[i, j, k] = alive     
        return(popsize)

    def make_bins(self, ts):
        locs = ts.individuals_location
        max_time = max(ts.individuals_time)
        for nx in np.arange(1, 20, 10):
            for ny in np.arange(1, 20, 10):
                for nt in np.arange(1, 60, 30):
                    yield [np.linspace(0, round(max(locs[:,0])), nx + 1), 
                           np.linspace(0, round(max(locs[:,1])), ny + 1),
                           np.round(np.linspace(0, max(nt, max_time), nt + 1))]  

    def verify(self, ts, remembered_stage):
        for bins in self.make_bins(ts):
            for stage in ('early', 'late'):
                x_bins, y_bins, time_bins = bins
                # as computed by pyslim
                popsize0 = pyslim.population_size(ts, x_bins, y_bins, time_bins, stage=stage, remembered_stage=remembered_stage)
                # as computed in a simple way
                popsize1 = self.population_size_simple(ts, x_bins, y_bins, time_bins, stage=stage, remembered_stage=remembered_stage)
                assert(np.allclose(popsize1, popsize0))

    @pytest.mark.parametrize('recipe', [next(recipe_eq("everyone"))], indirect=True)
    def test_errors(self, recipe):
        ts = recipe["ts"]
        x_bins = [0, 1.0]
        y_bins = [0, 1.0]
        time_bins = [0, 10.0]
        for stage in ['abcd', 10, []]:
            with pytest.raises(ValueError):
                pyslim.population_size(ts, x_bins, y_bins, time_bins, stage=stage)
            with pytest.raises(ValueError):
                pyslim.population_size(ts, x_bins, y_bins, time_bins, remembered_stage=stage)

    @pytest.mark.parametrize('recipe', [next(recipe_eq("pedigree", "WF"))], indirect=True)
    def test_mismatched_remembered_stage(self, recipe):
        ts = recipe["ts"]
        x_bins = [0, 1.0]
        y_bins = [0, 1.0]
        time_bins = [0, 10.0]
        info = recipe["info"]
        if "remembered_early" in recipe:
            with pytest.warns(UserWarning):
                pyslim.population_size(ts, x_bins, y_bins, time_bins, remembered_stage="late")
        else:
            with pytest.warns(UserWarning):
                pyslim.population_size(ts, x_bins, y_bins, time_bins, remembered_stage="early")

    @pytest.mark.skip("Waiting on next tskit release.")
    @pytest.mark.parametrize('recipe', recipe_eq("everyone"), indirect=True)
    def test_population_size(self, recipe):
        # compare output to the right answer
        ts = recipe["ts"]
        remembered_stage = 'early' if 'remembered_early' in recipe else 'late'
        self.verify(ts, remembered_stage=remembered_stage)

    def test_known_answer(self):
        # a simple example to make sure we've got the edge cases right
        tables = tskit.TableCollection(sequence_length=1)
        pyslim.set_tree_sequence_metadata(tables,  model_type='nonWF', tick=0)
        pyslim.set_metadata_schemas(tables)
        locs = [[0, 0], # alive at 0, 1
                [0, 1], # alive at 0, 1, 2
                [2, 0], # alive at 0, 1, 2
                [1, 1], # alive at 0, 1, 2
                [0, 0], # alive at 1
                [0.5, 1], # alive at 1
                [2, 2], # alive at 1
                [3, 2] # alive at 0, 1, 2, 3
        ]
        births = [1, 2, 2, 2, 1, 1, 1, 3]
        ages = [1, 2, 2, 2, 0, 0, 0, 3]
        x_bins = [0, 1, 3]
        for xy, a in zip(locs, ages):
            md = pyslim.default_slim_metadata('individual')
            md['age'] = a
            tables.individuals.add_row(
                    location=xy + [np.nan],
                    metadata=md,
            )

        for j, b in enumerate(births):
            tables.nodes.add_row(time=b, individual=j)

        ts = tables.tree_sequence()

        # check we've got this right
        for k, n in enumerate([
                [0, 1, 2, 3, 7],
                [0, 1, 2, 3, 4, 5, 6, 7],
                [1, 2, 3, 7],
                [7]]):
            np.testing.assert_array_equal(
                    n,
                    pyslim.individuals_alive_at(ts, k)
            )

        # no-one
        counts = pyslim.population_size(
                ts,
                x_bins=np.arange(10),
                y_bins=np.arange(10),
                time_bins=[100, 200, 300],
        )
        np.testing.assert_array_equal(
                counts,
                np.zeros((9, 9, 2)),
        )

        # everyone at the start
        counts = pyslim.population_size(
                ts,
                x_bins=[0, 10],
                y_bins=[0, 10],
                time_bins=[0, 1],
        )
        np.testing.assert_array_equal(
                counts,
                [[[5]]]
        )

        # should omit the last one
        counts = pyslim.population_size(
                ts,
                x_bins=[0, 3],
                y_bins=[0, 3],
                time_bins=[0, 1]
        )
        np.testing.assert_array_equal(
                counts,
                [[[4]]]
        )

        # now should omit the ones at the boundaries
        counts = pyslim.population_size(
                ts,
                x_bins=[0, 1, 2],
                y_bins=[0, 1, 2],
                time_bins=[0, 1, 2, 5]
        )
        np.testing.assert_array_equal(
                counts,
                [[[1, 2, 0],
                  [1, 2, 1/3]],
                 [[0, 0, 0],
                  [1, 1, 1/3]]]
        )



import msprime
import tskit
import json
import warnings
import numpy as np

from ._version import *
from .slim_metadata import *
from .provenance import *
from .util import *
from .slim_metadata import _old_metadata_schema

INDIVIDUAL_ALIVE = 2**16
"""
Used in ``individual.flags`` to denote the individual is alive
when the tree sequence was written out.
"""
INDIVIDUAL_REMEMBERED = 2**17
"""
Used in ``individual.flags`` to denote the individual was
marked as "remembered".
"""
INDIVIDUAL_RETAINED = 2**18
"""
Used in ``individual.flags`` to denote the individual was
marked as "retained".
"""
# deprecated but keep it around for backwards compatibility
# (also, it means effectively the same thing as RETAINED)
INDIVIDUAL_FIRST_GEN = INDIVIDUAL_RETAINED

NUCLEOTIDES = ['A', 'C', 'G', 'T']
"""
Mutation metadata records the nucleotide as an integer,
translated to ACGT by indexing this array,
so a nucleotide value of ``k`` actually means NUCLEOTIDES[k].
"""


def load(*args, **kwargs):
    raise RuntimeError("This method has been removed: use tskit.load( ) instead.")


def mutation_at(ts, node, position, time=None):
    '''
    Finds the mutation present in the genome of ``node`` at ``position``,
    returning -1 if there is no such mutation recorded in the tree
    sequence.  Warning: if ``node`` is not actually in the tree sequence
    (e.g., not ancestral to any samples) at ``position``, then this
    function will return -1, possibly erroneously.  If `time` is provided,
    returns the last mutation at ``position`` inherited by ``node`` that
    occurred at or before ``time`` ago.

    :param int node: The index of a node in the tree sequence.
    :param float position: A position along the genome.
    :param int time: The time ago that we want the nucleotide, or None,
        in which case the ``time`` of ``node`` is used.

    :returns: Index of the mutation in question, or -1 if none.
    '''
    if position < 0 or position >= ts.sequence_length:
        raise ValueError("Position {} not valid.".format(position))
    if node < 0 or node >= ts.num_nodes:
        raise ValueError("Node {} not valid.".format(node))
    if time is None:
        time = ts.node(node).time
    tree = ts.at(position)
    out = tskit.NULL
    try:
        site = ts.site(position=position)
    except ValueError:
        pass
    else:
        mut_nodes = []
        # look for only mutations that occurred before `time`
        # not strictly necessary if time was None
        for mut in site.mutations:
            if mut.time >= time:
                mut_nodes.append(mut.node)
        n = node
        while n > -1 and n not in mut_nodes:
            n = tree.parent(n)
        if n >= 0:
            # do careful error checking here
            for mut in site.mutations:
                if mut.node == n and mut.time >= time:
                    # BUG: this can fail if a mutation has two children
                    assert(out == tskit.NULL or out == mut.parent)
                    out = mut.id
    return out

def nucleotide_at(ts, node, position, time=None):
    '''
    Finds the nucleotide present in the genome of ``node`` at ``position``.
    Warning: if ``node`` is not actually in the tree sequence (e.g., not
    ancestral to any samples) at ``position``, then this function will
    return the reference sequence nucleotide, possibly erroneously.  If
    `time` is provided, returns the last nucletide produced by a mutation
    at ``position`` inherited by ``node`` that occurred at or before
    ``time`` ago.

    :param int node: The index of a node in the tree sequence.
    :param float position: A position along the genome.
    :param int time: The time ago that we want the nucleotide, or None,
        in which case the ``time`` of ``node`` is used.

    :returns: Index of the nucleotide in ``NUCLEOTIDES`` (0=A, 1=C, 2=G, 3=T).
    '''
    if not ts.has_reference_sequence():
        raise ValueError("This tree sequence has no reference sequence.")
    mut_id = mutation_at(ts, node, position, time)
    if mut_id == tskit.NULL:
        out = NUCLEOTIDES.index(ts.reference_sequence.data[int(position)])
    else:
        mut = ts.mutation(mut_id)
        k = np.argmax([u["slim_time"] for u in mut.metadata["mutation_list"]])
        out = mut.metadata["mutation_list"][k]["nucleotide"]
    return out
